"""Button platform for zhawss."""
from __future__ import annotations

import abc
import functools
from typing import TYPE_CHECKING, Any, Final, Type, Union

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_IDENTIFY

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.BUTTON)
DEFAULT_DURATION: Final[int] = 5  # seconds


class Button(PlatformEntity):
    """Button platform entity."""

    _command_name: str
    PLATFORM = Platform.BUTTON

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize button."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: ClusterHandler = cluster_handlers[0]

    @abc.abstractmethod
    def get_args(self) -> list[Any]:
        """Return the arguments to use in the command."""

    def get_state(self) -> dict:
        """Return the arguments to use in the command."""

    async def async_press(self) -> None:
        """Send out a update command."""
        command = getattr(self._cluster_handler, self._command_name)
        arguments = self.get_args()
        await command(*arguments)

    def to_json(self) -> dict:
        """Return a JSON representation of the button."""
        json = super().to_json()
        json["command"] = self._command_name
        return json


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IDENTIFY)
class IdentifyButton(Button):
    """Identify button platform entity."""

    _command_name = "identify"

    @classmethod
    def create_platform_entity(
        cls: Type[IdentifyButton],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Return a platform entity if it is a supported configuration, otherwise return None
        """
        if PLATFORM_ENTITIES.prevent_entity_creation(
            Platform.BUTTON, device.ieee, CLUSTER_HANDLER_IDENTIFY
        ):
            return None
        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    def get_args(self) -> list[Any]:
        """Return the arguments to use in the command."""

        return [DEFAULT_DURATION]



"""Mapping registries for zhawss."""

from typing import Final

from zhaws.server.zigbee.decorators import DictRegistry, SetRegistry

PHILLIPS_REMOTE_CLUSTER: Final[int] = 0xFC00
SMARTTHINGS_ACCELERATION_CLUSTER: Final[int] = 0xFC02
SMARTTHINGS_HUMIDITY_CLUSTER: Final[int] = 0xFC45
VOC_LEVEL_CLUSTER: Final[int] = 0x042E

BINDABLE_CLUSTERS = SetRegistry()
HANDLER_ONLY_CLUSTERS = SetRegistry()
CLIENT_CLUSTER_HANDLER_REGISTRY = DictRegistry()
CLUSTER_HANDLER_REGISTRY = DictRegistry()



"""Helper classes for zhaws.client."""
from __future__ import annotations

from typing import Any, Literal, Optional, Union, cast

from zhaws.client.client import Client
from zhaws.client.model.commands import (
    AddGroupMembersCommand,
    AlarmControlPanelArmAwayCommand,
    AlarmControlPanelArmHomeCommand,
    AlarmControlPanelArmNightCommand,
    AlarmControlPanelDisarmCommand,
    AlarmControlPanelTriggerCommand,
    ButtonPressCommand,
    ClientDisconnectCommand,
    ClientListenCommand,
    ClientListenRawZCLCommand,
    ClimateSetFanModeCommand,
    ClimateSetHvacModeCommand,
    ClimateSetPresetModeCommand,
    ClimateSetTemperatureCommand,
    CommandResponse,
    CoverCloseCommand,
    CoverOpenCommand,
    CoverSetPositionCommand,
    CoverStopCommand,
    CreateGroupCommand,
    FanSetPercentageCommand,
    FanSetPresetModeCommand,
    FanTurnOffCommand,
    FanTurnOnCommand,
    GetDevicesCommand,
    GetDevicesResponse,
    GetGroupsCommand,
    GroupsResponse,
    LightTurnOffCommand,
    LightTurnOnCommand,
    LockClearUserLockCodeCommand,
    LockDisableUserLockCodeCommand,
    LockEnableUserLockCodeCommand,
    LockLockCommand,
    LockSetUserLockCodeCommand,
    LockUnlockCommand,
    NumberSetValueCommand,
    PermitJoiningCommand,
    PermitJoiningResponse,
    PlatformEntityRefreshStateCommand,
    ReadClusterAttributesCommand,
    ReadClusterAttributesResponse,
    ReconfigureDeviceCommand,
    RemoveGroupMembersCommand,
    RemoveGroupsCommand,
    SelectSelectOptionCommand,
    SirenTurnOffCommand,
    SirenTurnOnCommand,
    StartNetworkCommand,
    StopNetworkCommand,
    StopServerCommand,
    SwitchTurnOffCommand,
    SwitchTurnOnCommand,
    UpdateGroupResponse,
    UpdateNetworkTopologyCommand,
    WriteClusterAttributeCommand,
    WriteClusterAttributeResponse,
)
from zhaws.client.model.types import (
    BaseEntity,
    BasePlatformEntity,
    Device,
    Group,
    GroupEntity,
)
from zhaws.server.platforms.registries import Platform


def ensure_platform_entity(entity: BaseEntity, platform: Platform) -> None:
    """Ensure an entity exists and is from the specified platform."""
    if entity is None or entity.platform != platform:
        raise ValueError(
            f"entity must be provided and it must be a {platform} platform entity"
        )


class LightHelper:
    """Helper to issue light commands."""

    def __init__(self, client: Client):
        """Initialize the light helper."""
        self._client: Client = client

    async def turn_on(
        self,
        light_platform_entity: BasePlatformEntity | GroupEntity,
        brightness: int | None = None,
        transition: int | None = None,
        flash: bool | None = None,
        effect: str | None = None,
        hs_color: tuple | None = None,
        color_temp: int | None = None,
    ) -> CommandResponse:
        """Turn on a light."""
        ensure_platform_entity(light_platform_entity, Platform.LIGHT)
        command = LightTurnOnCommand(
            ieee=light_platform_entity.device_ieee
            if not isinstance(light_platform_entity, GroupEntity)
            else None,
            group_id=light_platform_entity.group_id
            if isinstance(light_platform_entity, GroupEntity)
            else None,
            unique_id=light_platform_entity.unique_id,
            brightness=brightness,
            transition=transition,
            flash=flash,
            effect=effect,
            hs_color=hs_color,
            color_temp=color_temp,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def turn_off(
        self,
        light_platform_entity: BasePlatformEntity | GroupEntity,
        transition: int | None = None,
        flash: bool | None = None,
    ) -> CommandResponse:
        """Turn off a light."""
        ensure_platform_entity(light_platform_entity, Platform.LIGHT)
        command = LightTurnOffCommand(
            ieee=light_platform_entity.device_ieee
            if not isinstance(light_platform_entity, GroupEntity)
            else None,
            group_id=light_platform_entity.group_id
            if isinstance(light_platform_entity, GroupEntity)
            else None,
            unique_id=light_platform_entity.unique_id,
            transition=transition,
            flash=flash,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class SwitchHelper:
    """Helper to issue switch commands."""

    def __init__(self, client: Client):
        """Initialize the switch helper."""
        self._client: Client = client

    async def turn_on(
        self,
        switch_platform_entity: BasePlatformEntity | GroupEntity,
    ) -> CommandResponse:
        """Turn on a switch."""
        ensure_platform_entity(switch_platform_entity, Platform.SWITCH)
        command = SwitchTurnOnCommand(
            ieee=switch_platform_entity.device_ieee
            if not isinstance(switch_platform_entity, GroupEntity)
            else None,
            group_id=switch_platform_entity.group_id
            if isinstance(switch_platform_entity, GroupEntity)
            else None,
            unique_id=switch_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def turn_off(
        self,
        switch_platform_entity: BasePlatformEntity | GroupEntity,
    ) -> CommandResponse:
        """Turn off a switch."""
        ensure_platform_entity(switch_platform_entity, Platform.SWITCH)
        command = SwitchTurnOffCommand(
            ieee=switch_platform_entity.device_ieee
            if not isinstance(switch_platform_entity, GroupEntity)
            else None,
            group_id=switch_platform_entity.group_id
            if isinstance(switch_platform_entity, GroupEntity)
            else None,
            unique_id=switch_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class SirenHelper:
    """Helper to issue siren commands."""

    def __init__(self, client: Client):
        """Initialize the siren helper."""
        self._client: Client = client

    async def turn_on(
        self,
        siren_platform_entity: BasePlatformEntity,
        duration: Optional[int] = None,
        volume_level: Optional[int] = None,
        tone: Optional[str] = None,
    ) -> CommandResponse:
        """Turn on a siren."""
        ensure_platform_entity(siren_platform_entity, Platform.SIREN)
        command = SirenTurnOnCommand(
            ieee=siren_platform_entity.device_ieee,
            unique_id=siren_platform_entity.unique_id,
            duration=duration,
            volume_level=volume_level,
            tone=tone,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def turn_off(
        self, siren_platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Turn off a siren."""
        ensure_platform_entity(siren_platform_entity, Platform.SIREN)
        command = SirenTurnOffCommand(
            ieee=siren_platform_entity.device_ieee,
            unique_id=siren_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class ButtonHelper:
    """Helper to issue button commands."""

    def __init__(self, client: Client):
        """Initialize the button helper."""
        self._client: Client = client

    async def press(
        self, button_platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Press a button."""
        ensure_platform_entity(button_platform_entity, Platform.BUTTON)
        command = ButtonPressCommand(
            ieee=button_platform_entity.device_ieee,
            unique_id=button_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class CoverHelper:
    """helper to issue cover commands"""

    def __init__(self, client: Client):
        """Initialize the cover helper."""
        self._client: Client = client

    async def open_cover(
        self, cover_platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Open a cover."""
        ensure_platform_entity(cover_platform_entity, Platform.COVER)
        command = CoverOpenCommand(
            ieee=cover_platform_entity.device_ieee,
            unique_id=cover_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def close_cover(
        self, cover_platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Close a cover."""
        ensure_platform_entity(cover_platform_entity, Platform.COVER)
        command = CoverCloseCommand(
            ieee=cover_platform_entity.device_ieee,
            unique_id=cover_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def stop_cover(
        self, cover_platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Stop a cover."""
        ensure_platform_entity(cover_platform_entity, Platform.COVER)
        command = CoverStopCommand(
            ieee=cover_platform_entity.device_ieee,
            unique_id=cover_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_cover_position(
        self,
        cover_platform_entity: BasePlatformEntity,
        position: int,
    ) -> CommandResponse:
        """Set a cover position."""
        ensure_platform_entity(cover_platform_entity, Platform.COVER)
        command = CoverSetPositionCommand(
            ieee=cover_platform_entity.device_ieee,
            unique_id=cover_platform_entity.unique_id,
            position=position,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class FanHelper:
    """Helper to issue fan commands."""

    def __init__(self, client: Client):
        """Initialize the fan helper."""
        self._client: Client = client

    async def turn_on(
        self,
        fan_platform_entity: BasePlatformEntity | GroupEntity,
        speed: Optional[str] = None,
        percentage: Optional[int] = None,
        preset_mode: Optional[str] = None,
    ) -> CommandResponse:
        """Turn on a fan."""
        ensure_platform_entity(fan_platform_entity, Platform.FAN)
        command = FanTurnOnCommand(
            ieee=fan_platform_entity.device_ieee
            if not isinstance(fan_platform_entity, GroupEntity)
            else None,
            group_id=fan_platform_entity.group_id
            if isinstance(fan_platform_entity, GroupEntity)
            else None,
            unique_id=fan_platform_entity.unique_id,
            speed=speed,
            percentage=percentage,
            preset_mode=preset_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def turn_off(
        self,
        fan_platform_entity: BasePlatformEntity | GroupEntity,
    ) -> CommandResponse:
        """Turn off a fan."""
        ensure_platform_entity(fan_platform_entity, Platform.FAN)
        command = FanTurnOffCommand(
            ieee=fan_platform_entity.device_ieee
            if not isinstance(fan_platform_entity, GroupEntity)
            else None,
            group_id=fan_platform_entity.group_id
            if isinstance(fan_platform_entity, GroupEntity)
            else None,
            unique_id=fan_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_fan_percentage(
        self,
        fan_platform_entity: BasePlatformEntity | GroupEntity,
        percentage: int,
    ) -> CommandResponse:
        """Set a fan percentage."""
        ensure_platform_entity(fan_platform_entity, Platform.FAN)
        command = FanSetPercentageCommand(
            ieee=fan_platform_entity.device_ieee
            if not isinstance(fan_platform_entity, GroupEntity)
            else None,
            group_id=fan_platform_entity.group_id
            if isinstance(fan_platform_entity, GroupEntity)
            else None,
            unique_id=fan_platform_entity.unique_id,
            percentage=percentage,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_fan_preset_mode(
        self,
        fan_platform_entity: BasePlatformEntity | GroupEntity,
        preset_mode: str,
    ) -> CommandResponse:
        """Set a fan preset mode."""
        ensure_platform_entity(fan_platform_entity, Platform.FAN)
        command = FanSetPresetModeCommand(
            ieee=fan_platform_entity.device_ieee
            if not isinstance(fan_platform_entity, GroupEntity)
            else None,
            group_id=fan_platform_entity.group_id
            if isinstance(fan_platform_entity, GroupEntity)
            else None,
            unique_id=fan_platform_entity.unique_id,
            preset_mode=preset_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class LockHelper:
    """Helper to issue lock commands."""

    def __init__(self, client: Client):
        """Initialize the lock helper."""
        self._client: Client = client

    async def lock(self, lock_platform_entity: BasePlatformEntity) -> CommandResponse:
        """Lock a lock."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockLockCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def unlock(self, lock_platform_entity: BasePlatformEntity) -> CommandResponse:
        """Unlock a lock."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockUnlockCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_user_lock_code(
        self,
        lock_platform_entity: BasePlatformEntity,
        code_slot: int,
        user_code: str,
    ) -> CommandResponse:
        """Set a user lock code."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockSetUserLockCodeCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
            code_slot=code_slot,
            user_code=user_code,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def clear_user_lock_code(
        self,
        lock_platform_entity: BasePlatformEntity,
        code_slot: int,
    ) -> CommandResponse:
        """Clear a user lock code."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockClearUserLockCodeCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
            code_slot=code_slot,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def enable_user_lock_code(
        self,
        lock_platform_entity: BasePlatformEntity,
        code_slot: int,
    ) -> CommandResponse:
        """Enable a user lock code."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockEnableUserLockCodeCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
            code_slot=code_slot,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def disable_user_lock_code(
        self,
        lock_platform_entity: BasePlatformEntity,
        code_slot: int,
    ) -> CommandResponse:
        """Disable a user lock code."""
        ensure_platform_entity(lock_platform_entity, Platform.LOCK)
        command = LockDisableUserLockCodeCommand(
            ieee=lock_platform_entity.device_ieee,
            unique_id=lock_platform_entity.unique_id,
            code_slot=code_slot,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class NumberHelper:
    """Helper to issue number commands."""

    def __init__(self, client: Client):
        """Initialize the number helper."""
        self._client: Client = client

    async def set_value(
        self,
        number_platform_entity: BasePlatformEntity,
        value: Union[int, float],
    ) -> CommandResponse:
        """Set a number."""
        ensure_platform_entity(number_platform_entity, Platform.NUMBER)
        command = NumberSetValueCommand(
            ieee=number_platform_entity.device_ieee,
            unique_id=number_platform_entity.unique_id,
            value=value,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class SelectHelper:
    """Helper to issue select commands."""

    def __init__(self, client: Client):
        """Initialize the select helper."""
        self._client: Client = client

    async def select_option(
        self,
        select_platform_entity: BasePlatformEntity,
        option: Union[str, int],
    ) -> CommandResponse:
        """Set a select."""
        ensure_platform_entity(select_platform_entity, Platform.SELECT)
        command = SelectSelectOptionCommand(
            ieee=select_platform_entity.device_ieee,
            unique_id=select_platform_entity.unique_id,
            option=option,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class ClimateHelper:
    """Helper to issue climate commands."""

    def __init__(self, client: Client):
        """Initialize the climate helper."""
        self._client: Client = client

    async def set_hvac_mode(
        self,
        climate_platform_entity: BasePlatformEntity,
        hvac_mode: Literal[
            "heat_cool", "heat", "cool", "auto", "dry", "fan_only", "off"
        ],
    ) -> CommandResponse:
        """Set a climate."""
        ensure_platform_entity(climate_platform_entity, Platform.CLIMATE)
        command = ClimateSetHvacModeCommand(
            ieee=climate_platform_entity.device_ieee,
            unique_id=climate_platform_entity.unique_id,
            hvac_mode=hvac_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_temperature(
        self,
        climate_platform_entity: BasePlatformEntity,
        hvac_mode: Optional[
            Literal["heat_cool", "heat", "cool", "auto", "dry", "fan_only", "off"]
        ] = None,
        temperature: Optional[float] = None,
        target_temp_high: Optional[float] = None,
        target_temp_low: Optional[float] = None,
    ) -> CommandResponse:
        """Set a climate."""
        ensure_platform_entity(climate_platform_entity, Platform.CLIMATE)
        command = ClimateSetTemperatureCommand(
            ieee=climate_platform_entity.device_ieee,
            unique_id=climate_platform_entity.unique_id,
            temperature=temperature,
            target_temp_high=target_temp_high,
            target_temp_low=target_temp_low,
            hvac_mode=hvac_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_fan_mode(
        self,
        climate_platform_entity: BasePlatformEntity,
        fan_mode: str,
    ) -> CommandResponse:
        """Set a climate."""
        ensure_platform_entity(climate_platform_entity, Platform.CLIMATE)
        command = ClimateSetFanModeCommand(
            ieee=climate_platform_entity.device_ieee,
            unique_id=climate_platform_entity.unique_id,
            fan_mode=fan_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def set_preset_mode(
        self,
        climate_platform_entity: BasePlatformEntity,
        preset_mode: str,
    ) -> CommandResponse:
        """Set a climate."""
        ensure_platform_entity(climate_platform_entity, Platform.CLIMATE)
        command = ClimateSetPresetModeCommand(
            ieee=climate_platform_entity.device_ieee,
            unique_id=climate_platform_entity.unique_id,
            preset_mode=preset_mode,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class AlarmControlPanelHelper:
    """Helper to issue alarm control panel commands."""

    def __init__(self, client: Client):
        """Initialize the alarm control panel helper."""
        self._client: Client = client

    async def disarm(
        self, alarm_control_panel_platform_entity: BasePlatformEntity, code: str
    ) -> CommandResponse:
        """Disarm an alarm control panel."""
        ensure_platform_entity(
            alarm_control_panel_platform_entity, Platform.ALARM_CONTROL_PANEL
        )
        command = AlarmControlPanelDisarmCommand(
            ieee=alarm_control_panel_platform_entity.device_ieee,
            unique_id=alarm_control_panel_platform_entity.unique_id,
            code=code,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def arm_home(
        self, alarm_control_panel_platform_entity: BasePlatformEntity, code: str
    ) -> CommandResponse:
        """Arm an alarm control panel in home mode."""
        ensure_platform_entity(
            alarm_control_panel_platform_entity, Platform.ALARM_CONTROL_PANEL
        )
        command = AlarmControlPanelArmHomeCommand(
            ieee=alarm_control_panel_platform_entity.device_ieee,
            unique_id=alarm_control_panel_platform_entity.unique_id,
            code=code,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def arm_away(
        self, alarm_control_panel_platform_entity: BasePlatformEntity, code: str
    ) -> CommandResponse:
        """Arm an alarm control panel in away mode."""
        ensure_platform_entity(
            alarm_control_panel_platform_entity, Platform.ALARM_CONTROL_PANEL
        )
        command = AlarmControlPanelArmAwayCommand(
            ieee=alarm_control_panel_platform_entity.device_ieee,
            unique_id=alarm_control_panel_platform_entity.unique_id,
            code=code,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def arm_night(
        self, alarm_control_panel_platform_entity: BasePlatformEntity, code: str
    ) -> CommandResponse:
        """Arm an alarm control panel in night mode."""
        ensure_platform_entity(
            alarm_control_panel_platform_entity, Platform.ALARM_CONTROL_PANEL
        )
        command = AlarmControlPanelArmNightCommand(
            ieee=alarm_control_panel_platform_entity.device_ieee,
            unique_id=alarm_control_panel_platform_entity.unique_id,
            code=code,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def trigger(
        self,
        alarm_control_panel_platform_entity: BasePlatformEntity,
    ) -> CommandResponse:
        """Trigger an alarm control panel alarm."""
        ensure_platform_entity(
            alarm_control_panel_platform_entity, Platform.ALARM_CONTROL_PANEL
        )
        command = AlarmControlPanelTriggerCommand(
            ieee=alarm_control_panel_platform_entity.device_ieee,
            unique_id=alarm_control_panel_platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class PlatformEntityHelper:
    """Helper to send global platform entity commands."""

    def __init__(self, client: Client):
        """Initialize the platform entity helper."""
        self._client: Client = client

    async def refresh_state(
        self, platform_entity: BasePlatformEntity
    ) -> CommandResponse:
        """Refresh the state of a platform entity."""
        command = PlatformEntityRefreshStateCommand(
            ieee=platform_entity.device_ieee,
            unique_id=platform_entity.unique_id,
        )
        return await self._client.async_send_command(command.dict(exclude_none=True))


class ClientHelper:
    """Helper to send client specific commands."""

    def __init__(self, client: Client):
        """Initialize the client helper."""
        self._client: Client = client

    async def listen(self) -> CommandResponse:
        """Listen for incoming messages."""
        command = ClientListenCommand()
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def listen_raw_zcl(self) -> CommandResponse:
        """Listen for incoming raw ZCL messages."""
        command = ClientListenRawZCLCommand()
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def disconnect(self) -> CommandResponse:
        """Disconnect this client from the server."""
        command = ClientDisconnectCommand()
        return await self._client.async_send_command(command.dict(exclude_none=True))


class GroupHelper:
    """Helper to send group commands."""

    def __init__(self, client: Client):
        """Initialize the group helper."""
        self._client: Client = client

    async def get_groups(self) -> dict[int, Group]:
        """Get the groups."""
        response = cast(
            GroupsResponse,
            await self._client.async_send_command(
                GetGroupsCommand().dict(exclude_none=True)
            ),
        )
        return response.groups

    async def create_group(
        self,
        name: str,
        unique_id: Optional[int] = None,
        members: Optional[list[BasePlatformEntity]] = None,
    ) -> Group:
        """Create a new group."""
        request_data: dict[str, Any] = {
            "group_name": name,
            "group_id": unique_id,
        }
        if members is not None:
            request_data["members"] = [
                {"ieee": member.device_ieee, "endpoint_id": member.endpoint_id}
                for member in members
            ]

        command = CreateGroupCommand(**request_data)
        response = cast(
            UpdateGroupResponse,
            await self._client.async_send_command(command.dict(exclude_none=True)),
        )
        return response.group

    async def remove_groups(self, groups: list[Group]) -> dict[int, Group]:
        """Remove groups."""
        request: dict[str, Any] = {
            "group_ids": [group.id for group in groups],
        }
        command = RemoveGroupsCommand(**request)
        response = cast(
            GroupsResponse,
            await self._client.async_send_command(command.dict(exclude_none=True)),
        )
        return response.groups

    async def add_group_members(
        self, group: Group, members: list[BasePlatformEntity]
    ) -> Group:
        """Add members to a group."""
        request_data: dict[str, Any] = {
            "group_id": group.id,
            "members": [
                {"ieee": member.device_ieee, "endpoint_id": member.endpoint_id}
                for member in members
            ],
        }

        command = AddGroupMembersCommand(**request_data)
        response = cast(
            UpdateGroupResponse,
            await self._client.async_send_command(command.dict(exclude_none=True)),
        )
        return response.group

    async def remove_group_members(
        self, group: Group, members: list[BasePlatformEntity]
    ) -> Group:
        """Remove members from a group."""
        request_data: dict[str, Any] = {
            "group_id": group.id,
            "members": [
                {"ieee": member.device_ieee, "endpoint_id": member.endpoint_id}
                for member in members
            ],
        }

        command = RemoveGroupMembersCommand(**request_data)
        response = cast(
            UpdateGroupResponse,
            await self._client.async_send_command(command.dict(exclude_none=True)),
        )
        return response.group


class DeviceHelper:
    """Helper to send device commands."""

    def __init__(self, client: Client):
        """Initialize the device helper."""
        self._client: Client = client

    async def get_devices(self) -> dict[str, Device]:
        """Get the groups."""
        response = cast(
            GetDevicesResponse,
            await self._client.async_send_command(
                GetDevicesCommand().dict(exclude_none=True)
            ),
        )
        return response.devices

    async def reconfigure_device(self, device: Device) -> None:
        """Reconfigure a device."""
        await self._client.async_send_command(
            ReconfigureDeviceCommand(ieee=device.ieee).dict(exclude_none=True)
        )

    async def read_cluster_attributes(
        self,
        device: Device,
        cluster_id: int,
        cluster_type: str,
        endpoint_id: int,
        attributes: list[str],
        manufacturer_code: Optional[int] = None,
    ) -> ReadClusterAttributesResponse:
        """Read cluster attributes."""
        response = cast(
            ReadClusterAttributesResponse,
            await self._client.async_send_command(
                ReadClusterAttributesCommand(
                    ieee=device.ieee,
                    endpoint_id=endpoint_id,
                    cluster_id=cluster_id,
                    cluster_type=cluster_type,
                    attributes=attributes,
                    manufacturer_code=manufacturer_code,
                ).dict(exclude_none=True)
            ),
        )
        return response

    async def write_cluster_attribute(
        self,
        device: Device,
        cluster_id: int,
        cluster_type: str,
        endpoint_id: int,
        attribute: str,
        value: Any,
        manufacturer_code: Optional[int] = None,
    ) -> WriteClusterAttributeResponse:
        """Set the value for a cluster attribute."""
        response = cast(
            WriteClusterAttributeResponse,
            await self._client.async_send_command(
                WriteClusterAttributeCommand(
                    ieee=device.ieee,
                    endpoint_id=endpoint_id,
                    cluster_id=cluster_id,
                    cluster_type=cluster_type,
                    attribute=attribute,
                    value=value,
                    manufacturer_code=manufacturer_code,
                ).dict(exclude_none=True)
            ),
        )
        return response


class NetworkHelper:
    """Helper for network commands."""

    def __init__(self, client: Client):
        """Initialize the device helper."""
        self._client: Client = client

    async def permit_joining(
        self, duration: int = 255, device: Optional[Device] = None
    ) -> bool:
        """Permit joining for a specified duration."""
        # TODO add permit with code support
        request_data: dict[str, Any] = {
            "duration": duration,
        }
        if device is not None:
            if device.device_type == "EndDevice":
                raise ValueError("Device is not a coordinator or router")
            request_data["ieee"] = device.ieee
        command = PermitJoiningCommand(**request_data)
        response = cast(
            PermitJoiningResponse,
            await self._client.async_send_command(command.dict(exclude_none=True)),
        )
        return response.success

    async def update_topology(self) -> None:
        """Update the network topology."""
        await self._client.async_send_command(
            UpdateNetworkTopologyCommand().dict(exclude_none=True)
        )

    async def start_network(self) -> bool:
        """Start the Zigbee network."""
        command = StartNetworkCommand()
        response = await self._client.async_send_command(
            command.dict(exclude_none=True)
        )
        return response.success

    async def stop_network(self) -> bool:
        """Stop the Zigbee network."""
        response = await self._client.async_send_command(
            StopNetworkCommand().dict(exclude_none=True)
        )
        return response.success


class ServerHelper:
    """Helper for server commands."""

    def __init__(self, client: Client):
        """Initialize the helper."""
        self._client: Client = client

    async def stop_server(self) -> bool:
        """Stop the websocket server."""
        response = await self._client.async_send_command(
            StopServerCommand().dict(exclude_none=True)
        )
        return response.success



"""Decorators for zhawss."""

import asyncio
import logging
import random
from typing import Any, Callable, Coroutine, Tuple

_LOGGER = logging.getLogger(__name__)


def periodic(refresh_interval: Tuple) -> Callable:
    def scheduler(func: Callable) -> Callable[[Any, Any], Coroutine[Any, Any, None]]:
        async def wrapper(*args: Any, **kwargs: Any) -> None:
            sleep_time = random.randint(*refresh_interval)
            method_info = f"[{func.__module__}::{func.__qualname__}]"
            _LOGGER.debug(
                "Sleep time for periodic task: %s is %s seconds",
                method_info,
                sleep_time,
            )
            while True:
                try:
                    _LOGGER.debug(
                        "[%s] executing periodic task %s",
                        asyncio.current_task(),
                        method_info,
                    )
                    await func(*args, **kwargs)
                except asyncio.CancelledError:
                    _LOGGER.debug(
                        "[%s] Periodic task %s cancelled",
                        asyncio.current_task(),
                        method_info,
                    )
                    break
                except Exception:
                    _LOGGER.warning(
                        "[%s] Failed to poll using function %s",
                        asyncio.current_task(),
                        method_info,
                        exc_info=True,
                    )
                await asyncio.sleep(sleep_time)

        return wrapper

    return scheduler



"""Websocket application to run a zigpy Zigbee network."""
from __future__ import annotations

import argparse
import asyncio
import logging
from pathlib import Path

from zhaws.server.config.model import ServerConfiguration
from zhaws.server.websocket.server import Server

_LOGGER = logging.getLogger(__name__)


async def main(config_path: str | None = None) -> None:
    if config_path is None:
        _LOGGER.info("No config file provided, using default configuration")
        configuration = ServerConfiguration()
    else:
        _LOGGER.info("Loading configuration from %s", config_path)
        path = Path(config_path)
        configuration = ServerConfiguration.parse_file(path)
    async with Server(configuration=configuration) as server:
        await server.wait_closed()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start the ZHAWS server")
    parser.add_argument(
        "--config", type=str, default=None, help="Path to the configuration file"
    )

    args = parser.parse_args()

    import uvloop

    uvloop.install()

    from colorlog import ColoredFormatter

    fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"
    colorfmt = f"%(log_color)s{fmt}%(reset)s"
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger().handlers[0].setFormatter(
        ColoredFormatter(
            colorfmt,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        )
    )

    asyncio.run(main(args.config))



"""
HVAC cluster handlers module for zhawss.

For more details about this component, please refer to the documentation at
https://home-assistant.io/integrations/zha/
"""
from __future__ import annotations

from typing import Any

from zigpy.exceptions import ZigbeeException
from zigpy.zcl.clusters import hvac
from zigpy.zcl.foundation import Status

from zhaws.server.zigbee import registries
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
    ClusterHandler,
)
from zhaws.server.zigbee.cluster.const import (
    REPORT_CONFIG_MAX_INT,
    REPORT_CONFIG_MIN_INT,
    REPORT_CONFIG_OP,
)

REPORT_CONFIG_CLIMATE = (REPORT_CONFIG_MIN_INT, REPORT_CONFIG_MAX_INT, 25)
REPORT_CONFIG_CLIMATE_DEMAND = (REPORT_CONFIG_MIN_INT, REPORT_CONFIG_MAX_INT, 5)
REPORT_CONFIG_CLIMATE_DISCRETE = (REPORT_CONFIG_MIN_INT, REPORT_CONFIG_MAX_INT, 1)


@registries.CLUSTER_HANDLER_REGISTRY.register(hvac.Dehumidification.cluster_id)
class Dehumidification(ClusterHandler):
    """Dehumidification cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(hvac.Fan.cluster_id)
class FanClusterHandler(ClusterHandler):
    """Fan cluster handler."""

    _value_attribute = 0

    REPORT_CONFIG = [{"attr": "fan_mode", "config": REPORT_CONFIG_OP}]
    ZCL_INIT_ATTRS = {"fan_mode_sequence": True}

    @property
    def fan_mode(self) -> int | None:
        """Return current fan mode."""
        return self.cluster.get("fan_mode")

    @property
    def fan_mode_sequence(self) -> int | None:
        """Return possible fan mode speeds."""
        return self.cluster.get("fan_mode_sequence")

    async def async_set_speed(self, value: Any) -> None:
        """Set the speed of the fan."""

        try:
            await self.cluster.write_attributes({"fan_mode": value})
        except ZigbeeException as ex:
            self.error("Could not set speed: %s", ex)
            return

    async def async_update(self) -> None:
        """Retrieve latest state."""
        await self.get_attribute_value("fan_mode", from_cache=False)

    def attribute_updated(self, attrid: int, value: Any) -> None:
        """Handle attribute update from fan cluster."""
        attr_name = self.cluster.attributes.get(attrid, [attrid])[0]
        self.debug(
            "Attribute report '%s'[%s] = %s", self.cluster.name, attr_name, value
        )
        if attr_name == "fan_mode":
            self.emit(
                CLUSTER_HANDLER_EVENT,
                ClusterAttributeUpdatedEvent(
                    id=attrid,
                    name=attr_name,
                    value=value,
                ),
            )


@registries.CLUSTER_HANDLER_REGISTRY.register(hvac.Pump.cluster_id)
class Pump(ClusterHandler):
    """Pump cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(hvac.Thermostat.cluster_id)
class ThermostatClusterHandler(ClusterHandler):
    """Thermostat cluster handler."""

    REPORT_CONFIG = [
        {"attr": "local_temp", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "occupied_cooling_setpoint", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "occupied_heating_setpoint", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "unoccupied_cooling_setpoint", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "unoccupied_heating_setpoint", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "running_mode", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "running_state", "config": REPORT_CONFIG_CLIMATE_DEMAND},
        {"attr": "system_mode", "config": REPORT_CONFIG_CLIMATE},
        {"attr": "occupancy", "config": REPORT_CONFIG_CLIMATE_DISCRETE},
        {"attr": "pi_cooling_demand", "config": REPORT_CONFIG_CLIMATE_DEMAND},
        {"attr": "pi_heating_demand", "config": REPORT_CONFIG_CLIMATE_DEMAND},
    ]
    ZCL_INIT_ATTRS: dict[int | str, bool] = {
        "abs_min_heat_setpoint_limit": True,
        "abs_max_heat_setpoint_limit": True,
        "abs_min_cool_setpoint_limit": True,
        "abs_max_cool_setpoint_limit": True,
        "ctrl_seqe_of_oper": False,
        "max_cool_setpoint_limit": True,
        "max_heat_setpoint_limit": True,
        "min_cool_setpoint_limit": True,
        "min_heat_setpoint_limit": True,
    }

    @property
    def abs_max_cool_setpoint_limit(self) -> int:
        """Absolute maximum cooling setpoint."""
        return self.cluster.get("abs_max_cool_setpoint_limit", 3200)

    @property
    def abs_min_cool_setpoint_limit(self) -> int:
        """Absolute minimum cooling setpoint."""
        return self.cluster.get("abs_min_cool_setpoint_limit", 1600)

    @property
    def abs_max_heat_setpoint_limit(self) -> int:
        """Absolute maximum heating setpoint."""
        return self.cluster.get("abs_max_heat_setpoint_limit", 3000)

    @property
    def abs_min_heat_setpoint_limit(self) -> int:
        """Absolute minimum heating setpoint."""
        return self.cluster.get("abs_min_heat_setpoint_limit", 700)

    @property
    def ctrl_seqe_of_oper(self) -> int:
        """Control Sequence of operations attribute."""
        return self.cluster.get("ctrl_seqe_of_oper", 0xFF)

    @property
    def max_cool_setpoint_limit(self) -> int:
        """Maximum cooling setpoint."""
        sp_limit = self.cluster.get("max_cool_setpoint_limit")
        if sp_limit is None:
            return self.abs_max_cool_setpoint_limit
        return sp_limit

    @property
    def min_cool_setpoint_limit(self) -> int:
        """Minimum cooling setpoint."""
        sp_limit = self.cluster.get("min_cool_setpoint_limit")
        if sp_limit is None:
            return self.abs_min_cool_setpoint_limit
        return sp_limit

    @property
    def max_heat_setpoint_limit(self) -> int:
        """Maximum heating setpoint."""
        sp_limit = self.cluster.get("max_heat_setpoint_limit")
        if sp_limit is None:
            return self.abs_max_heat_setpoint_limit
        return sp_limit

    @property
    def min_heat_setpoint_limit(self) -> int:
        """Minimum heating setpoint."""
        sp_limit = self.cluster.get("min_heat_setpoint_limit")
        if sp_limit is None:
            return self.abs_min_heat_setpoint_limit
        return sp_limit

    @property
    def local_temp(self) -> int | None:
        """Thermostat temperature."""
        return self.cluster.get("local_temp")

    @property
    def occupancy(self) -> int | None:
        """Is occupancy detected."""
        return self.cluster.get("occupancy")

    @property
    def occupied_cooling_setpoint(self) -> int | None:
        """Temperature when room is occupied."""
        return self.cluster.get("occupied_cooling_setpoint")

    @property
    def occupied_heating_setpoint(self) -> int | None:
        """Temperature when room is occupied."""
        return self.cluster.get("occupied_heating_setpoint")

    @property
    def pi_cooling_demand(self) -> int:
        """Cooling demand."""
        return self.cluster.get("pi_cooling_demand")

    @property
    def pi_heating_demand(self) -> int:
        """Heating demand."""
        return self.cluster.get("pi_heating_demand")

    @property
    def running_mode(self) -> int | None:
        """Thermostat running mode."""
        return self.cluster.get("running_mode")

    @property
    def running_state(self) -> int | None:
        """Thermostat running state, state of heat, cool, fan relays."""
        return self.cluster.get("running_state")

    @property
    def system_mode(self) -> int | None:
        """System mode."""
        return self.cluster.get("system_mode")

    @property
    def unoccupied_cooling_setpoint(self) -> int | None:
        """Temperature when room is not occupied."""
        return self.cluster.get("unoccupied_cooling_setpoint")

    @property
    def unoccupied_heating_setpoint(self) -> int | None:
        """Temperature when room is not occupied."""
        return self.cluster.get("unoccupied_heating_setpoint")

    def attribute_updated(self, attrid: int, value: Any) -> None:
        """Handle attribute update cluster."""
        attr_name = self.cluster.attributes.get(attrid, [attrid])[0]
        self.debug(
            "Attribute report '%s'[%s] = %s", self.cluster.name, attr_name, value
        )
        self.emit(
            CLUSTER_HANDLER_EVENT,
            ClusterAttributeUpdatedEvent(
                id=attrid,
                name=attr_name,
                value=value,
            ),
        )

    async def async_set_operation_mode(self, mode: Any) -> bool:
        """Set Operation mode."""
        if not await self.write_attributes({"system_mode": mode}):
            self.debug("couldn't set '%s' operation mode", mode)
            return False

        self.debug("set system to %s", mode)
        return True

    async def async_set_heating_setpoint(
        self, temperature: int, is_away: bool = False
    ) -> bool:
        """Set heating setpoint."""
        if is_away:
            data = {"unoccupied_heating_setpoint": temperature}
        else:
            data = {"occupied_heating_setpoint": temperature}
        if not await self.write_attributes(data):
            self.debug("couldn't set heating setpoint")
            return False

        return True

    async def async_set_cooling_setpoint(
        self, temperature: int, is_away: bool = False
    ) -> bool:
        """Set cooling setpoint."""
        if is_away:
            data = {"unoccupied_cooling_setpoint": temperature}
        else:
            data = {"occupied_cooling_setpoint": temperature}
        if not await self.write_attributes(data):
            self.debug("couldn't set cooling setpoint")
            return False
        self.debug("set cooling setpoint to %s", temperature)
        return True

    async def get_occupancy(self) -> bool | None:
        """Get unreportable occupancy attribute."""
        try:
            res, fail = await self.cluster.read_attributes(["occupancy"])
            self.debug("read 'occupancy' attr, success: %s, fail: %s", res, fail)
            if "occupancy" not in res:
                return None
            return bool(self.occupancy)
        except ZigbeeException as ex:
            self.debug("Couldn't read 'occupancy' attribute: %s", ex)
            return None

    async def write_attributes(self, data: dict, **kwargs: Any) -> bool:
        """Write attributes helper."""
        try:
            res = await self.cluster.write_attributes(data, **kwargs)
        except ZigbeeException as exc:
            self.debug("couldn't write %s: %s", data, exc)
            return False

        self.debug("wrote %s attrs, Status: %s", data, res)
        return self.check_result(res)

    @staticmethod
    def check_result(res: Any) -> bool:
        """Normalize the result."""
        if not isinstance(res, list):
            return False

        return all(record.status == Status.SUCCESS for record in res[0])


@registries.CLUSTER_HANDLER_REGISTRY.register(hvac.UserInterface.cluster_id)
class UserInterface(ClusterHandler):
    """User interface (thermostat) cluster handler."""



"""Lock platform for zhawss."""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Final

from zigpy.zcl.foundation import Status

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_DOORLOCK

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.LOCK)

STATE_LOCKED: Final[str] = "locked"
STATE_UNLOCKED: Final[str] = "unlocked"
STATE_LOCKING: Final[str] = "locking"
STATE_UNLOCKING: Final[str] = "unlocking"
STATE_JAMMED: Final[str] = "jammed"
# The first state is Zigbee 'Not fully locked'
STATE_LIST: Final[list[str]] = [STATE_UNLOCKED, STATE_LOCKED, STATE_UNLOCKED]
VALUE_TO_STATE: Final = dict(enumerate(STATE_LIST))


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_DOORLOCK)
class Lock(PlatformEntity):
    """Representation of a zhawss lock."""

    PLATFORM = Platform.LOCK

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the lock."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._doorlock_cluster_handler: ClusterHandler = cluster_handlers[0]
        self._state = VALUE_TO_STATE.get(
            self._doorlock_cluster_handler.cluster.get("lock_state"), None
        )
        self._doorlock_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def is_locked(self) -> bool:
        """Return true if entity is locked."""
        if self._state is None:
            return False
        return self._state == STATE_LOCKED

    async def async_lock(self, **kwargs: Any) -> None:
        """Lock the lock."""
        result = await self._doorlock_cluster_handler.lock_door()
        if not isinstance(result, list) or result[0] is not Status.SUCCESS:
            self.error("Error with lock_door: %s", result)
            return
        self.maybe_send_state_changed_event()

    async def async_unlock(self, **kwargs: Any) -> None:
        """Unlock the lock."""
        result = await self._doorlock_cluster_handler.unlock_door()
        if not isinstance(result, list) or result[0] is not Status.SUCCESS:
            self.error("Error with unlock_door: %s", result)
            return
        self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve state from the lock."""
        await super().async_update()
        await self.async_get_state()

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle state update from cluster handler."""
        self._state = VALUE_TO_STATE.get(event.value, self._state)
        self.maybe_send_state_changed_event()

    async def async_get_state(self, from_cache: bool = True) -> None:
        """Attempt to retrieve state from the lock."""
        if self._doorlock_cluster_handler:
            state = await self._doorlock_cluster_handler.get_attribute_value(
                "lock_state", from_cache=from_cache
            )
            if state is not None:
                self._state = VALUE_TO_STATE.get(state, self._state)

    async def async_set_lock_user_code(self, code_slot: int, user_code: str) -> None:
        """Set the user_code to index X on the lock."""
        if self._doorlock_cluster_handler:
            await self._doorlock_cluster_handler.async_set_user_code(
                code_slot, user_code
            )
            self.debug("User code at slot %s set", code_slot)

    async def async_enable_lock_user_code(self, code_slot: int) -> None:
        """Enable user_code at index X on the lock."""
        if self._doorlock_cluster_handler:
            await self._doorlock_cluster_handler.async_enable_user_code(code_slot)
            self.debug("User code at slot %s enabled", code_slot)

    async def async_disable_lock_user_code(self, code_slot: int) -> None:
        """Disable user_code at index X on the lock."""
        if self._doorlock_cluster_handler:
            await self._doorlock_cluster_handler.async_disable_user_code(code_slot)
            self.debug("User code at slot %s disabled", code_slot)

    async def async_clear_lock_user_code(self, code_slot: int) -> None:
        """Clear the user_code at index X on the lock."""
        if self._doorlock_cluster_handler:
            await self._doorlock_cluster_handler.async_clear_user_code(code_slot)
            self.debug("User code at slot %s cleared", code_slot)

    def get_state(self) -> dict:
        """Get the state of the lock."""
        response = super().get_state()
        response["is_locked"] = self.is_locked
        return response



"""WS api for the climate platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.platforms.climate import (
    ATTR_FAN_MODE,
    ATTR_HVAC_MODE,
    ATTR_TARGET_TEMP_HIGH,
    ATTR_TARGET_TEMP_LOW,
    ATTR_TEMPERATURE,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

# All activity disabled / Device is off/standby
HVAC_MODE_OFF = "off"

# Heating
HVAC_MODE_HEAT = "heat"

# Cooling
HVAC_MODE_COOL = "cool"

# The device supports heating/cooling to a range
HVAC_MODE_HEAT_COOL = "heat_cool"

# The temperature is set based on a schedule, learned behavior, AI or some
# other related mechanism. User is not able to adjust the temperature
HVAC_MODE_AUTO = "auto"

# Device is in Dry/Humidity mode
HVAC_MODE_DRY = "dry"

# Only the fan is on, not fan and another mode like cool
HVAC_MODE_FAN_ONLY = "fan_only"

HVAC_MODES = [
    HVAC_MODE_OFF,
    HVAC_MODE_HEAT,
    HVAC_MODE_COOL,
    HVAC_MODE_HEAT_COOL,
    HVAC_MODE_AUTO,
    HVAC_MODE_DRY,
    HVAC_MODE_FAN_ONLY,
]

ATTR_PRESET_MODE = "preset_mode"


class ClimateCommands(StrEnum):
    """Climate commands."""

    SET_FAN_MODE = "climate_set_fan_mode"
    SET_HVAC_MODE = "climate_set_hvac_mode"
    SET_PRESET_MODE = "climate_set_preset_mode"
    SET_TEMPERATURE = "climate_set_temperature"


@decorators.websocket_command(
    platform_entity_command_schema(
        ClimateCommands.SET_FAN_MODE,
        {
            vol.Required(ATTR_FAN_MODE): str,
        },
    )
)
@decorators.async_response
async def set_fan_mode(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Set the fan mode for the climate platform entity."""
    await execute_platform_entity_command(server, client, message, "async_set_fan_mode")


@decorators.websocket_command(
    platform_entity_command_schema(
        ClimateCommands.SET_HVAC_MODE,
        {
            vol.Required(ATTR_HVAC_MODE): vol.In(HVAC_MODES),
        },
    )
)
@decorators.async_response
async def set_hvac_mode(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the hvac mode for the climate platform entity."""
    await execute_platform_entity_command(
        server, client, message, "async_set_hvac_mode"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        ClimateCommands.SET_PRESET_MODE,
        {
            vol.Required(ATTR_PRESET_MODE): str,
        },
    )
)
@decorators.async_response
async def set_preset_mode(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the preset mode for the climate platform entity."""
    await execute_platform_entity_command(
        server, client, message, "async_set_preset_mode"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        ClimateCommands.SET_TEMPERATURE,
        {
            vol.Exclusive(ATTR_TEMPERATURE, "temperature"): vol.Coerce(float),
            vol.Inclusive(ATTR_TARGET_TEMP_HIGH, "temperature"): vol.Coerce(float),
            vol.Inclusive(ATTR_TARGET_TEMP_LOW, "temperature"): vol.Coerce(float),
            vol.Optional(ATTR_HVAC_MODE): vol.In(HVAC_MODES),
        },
    )
)
@decorators.async_response
async def set_temperature(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the temperature and hvac mode for the climate platform entity."""
    await execute_platform_entity_command(
        server, client, message, "async_set_temperature"
    )


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, set_fan_mode)
    register_api_command(server, set_hvac_mode)
    register_api_command(server, set_preset_mode)
    register_api_command(server, set_temperature)



"""Switch platform for zhawss."""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, cast

from zigpy.zcl.clusters.general import OnOff
from zigpy.zcl.foundation import Status

from zhaws.server.platforms import BaseEntity, GroupEntity, PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_ON_OFF
from zhaws.server.zigbee.cluster.general import OnOffClusterHandler
from zhaws.server.zigbee.group import Group

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.SWITCH)
GROUP_MATCH = functools.partial(PLATFORM_ENTITIES.group_match, Platform.SWITCH)


class BaseSwitch(BaseEntity):
    """Common base class for zhawss switches."""

    PLATFORM = Platform.SWITCH

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize the switch."""
        self._on_off_cluster_handler: OnOffClusterHandler
        self._state: bool | None = None
        super().__init__(*args, **kwargs)

    @property
    def is_on(self) -> bool:
        """Return if the switch is on based on the statemachine."""
        if self._state is None:
            return False
        return self._state

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the entity on."""
        result = await self._on_off_cluster_handler.on()
        if not isinstance(result, list) or result[1] is not Status.SUCCESS:
            return
        self._state = True
        self.maybe_send_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn the entity off."""
        result = await self._on_off_cluster_handler.off()
        if not isinstance(result, list) or result[1] is not Status.SUCCESS:
            return
        self._state = False
        self.maybe_send_state_changed_event()

    def get_state(self) -> dict:
        """Return the state of the switch."""
        response = super().get_state()
        response["state"] = self.is_on
        return response


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_ON_OFF)
class Switch(PlatformEntity, BaseSwitch):
    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the switch."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._on_off_cluster_handler: OnOffClusterHandler = cast(
            OnOffClusterHandler, self.cluster_handlers[CLUSTER_HANDLER_ON_OFF]
        )
        self._state: bool = bool(self._on_off_cluster_handler.on_off)
        self._on_off_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle state update from cluster handler."""
        self._state = bool(event.value)
        self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve on off state from the switch."""
        await super().async_update()
        if self._on_off_cluster_handler:
            state = await self._on_off_cluster_handler.get_attribute_value("on_off")
            if state is not None:
                self._state = state
                self.maybe_send_state_changed_event()


@GROUP_MATCH()
class SwitchGroup(GroupEntity, BaseSwitch):
    """Representation of a switch group."""

    def __init__(self, group: Group):
        """Initialize a switch group."""
        super().__init__(group)
        self._on_off_cluster_handler = group.zigpy_group.endpoint[OnOff.cluster_id]

    def update(self, _: Any | None = None) -> None:
        """Query all members and determine the light group state."""
        self.debug("Updating switch group entity state")
        platform_entities = self._group.get_platform_entities(self.PLATFORM)
        all_entities = [entity.to_json() for entity in platform_entities]
        all_states = [entity["state"] for entity in all_entities]
        self.debug(
            "All platform entity states for group entity members: %s", all_states
        )
        on_states = [state for state in all_states if state["state"]]

        self._state = len(on_states) > 0
        self._available = any(entity.available for entity in platform_entities)

        self.maybe_send_state_changed_event()



"""Configuration models for the zhaws server."""
from __future__ import annotations

from typing import Annotated, Literal, Union

from pydantic import Field

from zhaws.model import BaseModel


class BaseRadioConfiguration(BaseModel):
    """Base zigbee radio configuration for zhaws."""

    type: Literal["ezsp", "xbee", "deconz", "zigate", "znp"]
    path: str = "/dev/tty.SLAB_USBtoUART"


class EZSPRadioConfiguration(BaseRadioConfiguration):
    """EZSP radio configuration for zhaws."""

    type: Literal["ezsp"] = "ezsp"
    baudrate: int = 115200
    flow_control: Literal["hardware", "software"] = "hardware"


class XBeeRadioConfiguration(BaseRadioConfiguration):
    """XBee radio configuration for zhaws."""

    type: Literal["xbee"] = "xbee"


class DeconzRadioConfiguration(BaseRadioConfiguration):
    """Deconz radio configuration for zhaws."""

    type: Literal["deconz"] = "deconz"


class ZigateRadioConfiguration(BaseRadioConfiguration):
    """Zigate radio configuration for zhaws."""

    type: Literal["zigate"] = "zigate"


class ZNPRadioConfiguration(BaseRadioConfiguration):
    """ZNP radio configuration for zhaws."""

    type: Literal["znp"] = "znp"


class ZigpyConfiguration(BaseModel):
    """Zigpy configuration for zhaws."""

    database_path: str = "./zigbee.db"
    enable_quirks: bool = True


class ServerConfiguration(BaseModel):
    """Server configuration for zhaws."""

    host: str = "0.0.0.0"
    port: int = 8001
    network_auto_start: bool = False
    zigpy_configuration: ZigpyConfiguration = ZigpyConfiguration()
    radio_configuration: Annotated[
        Union[
            EZSPRadioConfiguration,
            XBeeRadioConfiguration,
            DeconzRadioConfiguration,
            ZigateRadioConfiguration,
            ZNPRadioConfiguration,
        ],
        Field(discriminator="type"),  # noqa: F821
    ] = EZSPRadioConfiguration()



"""Cover platform for zhawss."""
from __future__ import annotations

import asyncio
import functools
import logging
from typing import TYPE_CHECKING, Any, Final, Union

from zigpy.zcl.foundation import Status

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_COVER,
    CLUSTER_HANDLER_LEVEL,
    CLUSTER_HANDLER_ON_OFF,
    CLUSTER_HANDLER_SHADE,
)
from zhaws.server.zigbee.cluster.general import LevelChangeEvent

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.COVER)

ATTR_CURRENT_POSITION: Final[str] = "current_position"
ATTR_CURRENT_TILT_POSITION: Final[str] = "current_tilt_position"
ATTR_POSITION: Final[str] = "position"
ATTR_TILT_POSITION: Final[str] = "tilt_position"

STATE_OPEN: Final[str] = "open"
STATE_OPENING: Final[str] = "opening"
STATE_CLOSED: Final[str] = "closed"
STATE_CLOSING: Final[str] = "closing"

_LOGGER = logging.getLogger(__name__)


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_COVER)
class Cover(PlatformEntity):
    """Representation of a zhawss cover."""

    PLATFORM = Platform.COVER

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the cover."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cover_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_COVER
        ]
        self._current_position = None
        self._state = None
        if (
            self._cover_cluster_handler
        ):  # TODO this should brobably be changed to raise if None
            self._current_position = 100 - self._cover_cluster_handler.cluster.get(
                "current_position_lift_percentage"
            )
            if self._current_position == 0:
                self._state = STATE_CLOSED
            elif self._current_position == 100:
                self._state = STATE_OPEN
        self._cover_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def is_closed(self) -> bool | None:
        """Return if the cover is closed."""
        if self.current_cover_position is None:
            return None
        return self.current_cover_position == 0

    @property
    def is_opening(self) -> bool:
        """Return if the cover is opening or not."""
        return self._state == STATE_OPENING

    @property
    def is_closing(self) -> bool:
        """Return if the cover is closing or not."""
        return self._state == STATE_CLOSING

    @property
    def current_cover_position(self) -> int | None:
        """Return the current position of ZHA cover.
        None is unknown, 0 is closed, 100 is fully open.
        """
        return self._current_position

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle position update from cluster handler."""
        _LOGGER.debug("setting position: %s", event.value)
        self._current_position = 100 - event.value
        if self._current_position == 0:
            self._state = STATE_CLOSED
        elif self._current_position == 100:
            self._state = STATE_OPEN
        self.maybe_send_state_changed_event()

    def async_update_state(self, state: Any) -> None:
        """Handle state update from cluster handler."""
        _LOGGER.debug("state=%s", state)
        self._state = state
        self.maybe_send_state_changed_event()

    async def async_open_cover(self, **kwargs: Any) -> None:
        """Open the window cover."""
        res = await self._cover_cluster_handler.up_open()
        if isinstance(res, list) and res[1] is Status.SUCCESS:
            self.async_update_state(STATE_OPENING)

    async def async_close_cover(self, **kwargs: Any) -> None:
        """Close the window cover."""
        res = await self._cover_cluster_handler.down_close()
        if isinstance(res, list) and res[1] is Status.SUCCESS:
            self.async_update_state(STATE_CLOSING)

    async def async_set_cover_position(self, **kwargs: Any) -> None:
        """Move the roller shutter to a specific position."""
        new_pos = kwargs[ATTR_POSITION]
        res = await self._cover_cluster_handler.go_to_lift_percentage(100 - new_pos)
        if isinstance(res, list) and res[1] is Status.SUCCESS:
            self.async_update_state(
                STATE_CLOSING if new_pos < self._current_position else STATE_OPENING
            )

    async def async_stop_cover(self, **kwargs: Any) -> None:
        """Stop the window cover."""
        res = await self._cover_cluster_handler.stop()
        if isinstance(res, list) and res[1] is Status.SUCCESS:
            self._state = (
                STATE_OPEN
                if self._current_position is not None and self._current_position > 0
                else STATE_CLOSED
            )
            self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve the open/close state of the cover."""
        await super().async_update()
        await self.async_get_state()

    async def async_get_state(self, from_cache: bool = True) -> None:
        """Fetch the current state."""
        _LOGGER.debug("polling current state")
        if self._cover_cluster_handler:
            pos = await self._cover_cluster_handler.get_attribute_value(
                "current_position_lift_percentage", from_cache=from_cache
            )
            _LOGGER.debug("read pos=%s", pos)

            if pos is not None:
                self._current_position = 100 - pos
                self._state = (
                    STATE_OPEN
                    if self.current_cover_position is not None
                    and self.current_cover_position > 0
                    else STATE_CLOSED
                )
            else:
                self._current_position = None
                self._state = None

    def get_state(self) -> dict:
        """Get the state of the cover."""
        response = super().get_state()
        response.update(
            {
                ATTR_CURRENT_POSITION: self.current_cover_position,
                "state": self._state,
                "is_opening": self.is_opening,
                "is_closing": self.is_closing,
                "is_closed": self.is_closed,
            }
        )
        return response


@MULTI_MATCH(
    cluster_handler_names={
        CLUSTER_HANDLER_LEVEL,
        CLUSTER_HANDLER_ON_OFF,
        CLUSTER_HANDLER_SHADE,
    }
)
class Shade(PlatformEntity):
    """ZHAWSS Shade."""

    PLATFORM = Platform.COVER

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the cover."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._on_off_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_ON_OFF
        ]
        self._level_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_LEVEL
        ]
        self._is_open: bool = bool(self._on_off_cluster_handler.on_off)
        position = self._level_cluster_handler.current_level
        position = max(0, min(255, position))
        self._position: Union[int, None] = int(position * 100 / 255)
        self._on_off_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )
        self._level_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def current_cover_position(self) -> int | None:
        """Return current position of cover.
        None is unknown, 0 is closed, 100 is fully open.
        """
        return self._position

    @property
    def is_closed(self) -> Union[bool, None]:
        """Return True if shade is closed."""
        return not self._is_open

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Set open/closed state."""
        self._is_open = bool(event.value)
        self.maybe_send_state_changed_event()

    def handle_cluster_handler_set_level(self, event: LevelChangeEvent) -> None:
        """Set the reported position."""
        value = max(0, min(255, event.level))
        self._position = int(value * 100 / 255)
        self.maybe_send_state_changed_event()

    async def async_open_cover(self, **kwargs: Any) -> None:
        """Open the window cover."""
        res = await self._on_off_cluster_handler.on()
        if not isinstance(res, list) or res[1] != Status.SUCCESS:
            self.debug("couldn't open cover: %s", res)
            return

        self._is_open = True
        self.maybe_send_state_changed_event()

    async def async_close_cover(self, **kwargs: Any) -> None:
        """Close the window cover."""
        res = await self._on_off_cluster_handler.off()
        if not isinstance(res, list) or res[1] != Status.SUCCESS:
            self.debug("couldn't open cover: %s", res)
            return

        self._is_open = False
        self.maybe_send_state_changed_event()

    async def async_set_cover_position(self, position: int, **kwargs: Any) -> None:
        """Move the roller shutter to a specific position."""
        res = await self._level_cluster_handler.move_to_level_with_on_off(
            position * 255 / 100, 1
        )

        if not isinstance(res, list) or res[1] != Status.SUCCESS:
            self.debug("couldn't set cover's position: %s", res)
            return

        self._position = position
        self.maybe_send_state_changed_event()

    async def async_stop_cover(self, **kwargs: Any) -> None:
        """Stop the cover."""
        res = await self._level_cluster_handler.stop()
        if not isinstance(res, list) or res[1] != Status.SUCCESS:
            self.debug("couldn't stop cover: %s", res)
            return

    def get_state(self) -> dict:
        """Get the state of the cover."""
        response = super().get_state()
        response.update(
            {
                ATTR_CURRENT_POSITION: self._position,
                "is_closed": self.is_closed,
            }
        )
        return response


@MULTI_MATCH(
    cluster_handler_names={CLUSTER_HANDLER_LEVEL, CLUSTER_HANDLER_ON_OFF},
    manufacturers="Keen Home Inc",
)
class KeenVent(Shade):
    """Keen vent cover."""

    async def async_open_cover(self, **kwargs: Any) -> None:
        """Open the cover."""
        position = self._position or 100
        tasks = [
            self._level_cluster_handler.move_to_level_with_on_off(
                position * 255 / 100, 1
            ),
            self._on_off_cluster_handler.on(),
        ]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        if any(isinstance(result, Exception) for result in results):
            self.debug("couldn't open cover")
            return

        self._is_open = True
        self._position = position
        self.maybe_send_state_changed_event()



"""Models that represent commands and command responses."""

from typing import Annotated, Any, Literal, Optional, Tuple, Type, Union

from pydantic import conint
from pydantic.fields import Field

from zhaws.client.model.events import MinimalCluster, MinimalDevice
from zhaws.client.model.types import Device, Group, GroupMemberReference
from zhaws.model import BaseModel


class Command(BaseModel):
    """Base class for command that are sent to the server."""

    command: str


class PermitJoiningCommand(Command):
    """Command to permit joining."""

    command: Literal["permit_joining"] = "permit_joining"
    duration: Type[int] = conint(ge=1, le=254)
    device: Optional[Device]


class StopNetworkCommand(Command):
    """Command to stop the zigbee network."""

    command: Literal["stop_network"] = "stop_network"


class ZigbeeCoordinatorDeviceConfiguration(BaseModel):
    """Configuration of the zigbee coordinator."""

    path: str
    flow_control: str
    baudrate: int


class StartNetworkCommand(Command):
    """Command to start the zigbee network."""

    command: Literal["start_network"] = "start_network"


class StopServerCommand(Command):
    """Command to stop the websocket server."""

    command: Literal["stop_server"] = "stop_server"


class UpdateNetworkTopologyCommand(Command):
    """Command to update the network topology."""

    command: Literal["update_network_topology"] = "update_network_topology"


class GetDevicesCommand(Command):
    """Command to get devices."""

    command: Literal["get_devices"] = "get_devices"


class ReconfigureDeviceCommand(Command):
    """Command to reconfigure a device."""

    command: Literal["reconfigure_device"] = "reconfigure_device"
    ieee: str


class ReadClusterAttributesCommand(Command):
    """Command to get the values of the specified attributes."""

    command: Literal["read_cluster_attributes"] = "read_cluster_attributes"
    ieee: str
    endpoint_id: int
    cluster_id: int
    cluster_type: Literal["in", "out"]
    attributes: list[str]
    manufacturer_code: Optional[int]


class WriteClusterAttributeCommand(Command):
    """Command to set the value for a cluster attribute."""

    command: Literal["write_cluster_attribute"] = "write_cluster_attribute"
    ieee: str
    endpoint_id: int
    cluster_id: int
    cluster_type: Literal["in", "out"]
    attribute: str
    value: Any
    manufacturer_code: Optional[int]


class GetGroupsCommand(Command):
    """Command to get groups."""

    command: Literal["get_groups"] = "get_groups"


class CreateGroupCommand(Command):
    """Command to create a group."""

    command: Literal["create_group"] = "create_group"
    group_name: str
    group_id: Optional[int]
    members: Optional[list[GroupMemberReference]]


class RemoveGroupsCommand(Command):
    """Command to remove groups."""

    command: Literal["remove_groups"] = "remove_groups"
    group_ids: list[int]


class AddGroupMembersCommand(Command):
    """Command to add members to a group."""

    command: Literal["add_group_members"] = "add_group_members"
    group_id: int
    members: list[GroupMemberReference]


class RemoveGroupMembersCommand(Command):
    """Command to remove members from a group."""

    command: Literal["remove_group_members"] = "remove_group_members"
    group_id: int
    members: list[GroupMemberReference]


class PlatformEntityCommand(Command):
    """Base class for commands that address individual platform entities."""

    unique_id: str
    group_id: Optional[int]
    ieee: Optional[str]


class LightTurnOnCommand(PlatformEntityCommand):
    """Command to instruct a light to turn on."""

    command: Literal["light_turn_on"] = "light_turn_on"
    brightness: Optional[int]
    transition: Optional[int]
    flash: Optional[Literal["long", "short"]]
    effect: Optional[str]
    hs_color: Optional[Tuple[float, float]]
    color_temp: Optional[int]


class LightTurnOffCommand(PlatformEntityCommand):
    """Command to instruct a light to turn off."""

    command: Literal["light_turn_off"] = "light_turn_off"
    transition: Optional[int]
    flash: Optional[Literal["long", "short"]]


class SwitchTurnOnCommand(PlatformEntityCommand):
    """Command to instruct a switch to turn on."""

    command: Literal["switch_turn_on"] = "switch_turn_on"


class SwitchTurnOffCommand(PlatformEntityCommand):
    """Command to instruct a switch to turn off."""

    command: Literal["switch_turn_off"] = "switch_turn_off"


class SirenTurnOffCommand(PlatformEntityCommand):
    """Command to instruct a siren to turn off."""

    command: Literal["siren_turn_off"] = "siren_turn_off"


class SelectSelectOptionCommand(PlatformEntityCommand):
    """Command to instruct a select an option for a select platform entity."""

    command: Literal["select_select_option"] = "select_select_option"
    option: Union[str, int]


class NumberSetValueCommand(PlatformEntityCommand):
    """Command to set the value of a number platform entity."""

    command: Literal["number_set_value"] = "number_set_value"
    value: Union[float, int]


class SirenTurnOnCommand(PlatformEntityCommand):
    """Command to instruct a siren to turn off."""

    command: Literal["siren_turn_on"] = "siren_turn_on"
    duration: Optional[int]
    volume_level: Optional[int]
    tone: Optional[str]


class LockLockCommand(PlatformEntityCommand):
    """Command to lock a lock platform entity."""

    command: Literal["lock_lock"] = "lock_lock"


class LockUnlockCommand(PlatformEntityCommand):
    """Command to unlock a lock platform entity."""

    command: Literal["lock_unlock"] = "lock_unlock"


class LockDisableUserLockCodeCommand(PlatformEntityCommand):
    """Command to lock a lock platform entity."""

    command: Literal["lock_disable_user_lock_code"] = "lock_disable_user_lock_code"
    code_slot: int


class LockSetUserLockCodeCommand(PlatformEntityCommand):
    """Command to lock a lock platform entity."""

    command: Literal["lock_set_user_lock_code"] = "lock_set_user_lock_code"
    code_slot: int
    user_code: str


class LockEnableUserLockCodeCommand(PlatformEntityCommand):
    """Command to lock a lock platform entity."""

    command: Literal["lock_enable_user_lock_code"] = "lock_enable_user_lock_code"
    code_slot: int


class LockClearUserLockCodeCommand(PlatformEntityCommand):
    """Command to lock a lock platform entity."""

    command: Literal["lock_clear_user_lock_code"] = "lock_clear_user_lock_code"
    code_slot: int


class FanTurnOnCommand(PlatformEntityCommand):
    """Command to turn on a fan platform entity."""

    command: Literal["fan_turn_on"] = "fan_turn_on"
    speed: Optional[str]
    percentage: Optional[conint(ge=0, le=100)]  # type: ignore #TODO see if there is a way to make this work mypy
    preset_mode: Optional[str]


class FanTurnOffCommand(PlatformEntityCommand):
    """Command to turn off a fan platform entity."""

    command: Literal["fan_turn_off"] = "fan_turn_off"


class FanSetPercentageCommand(PlatformEntityCommand):
    """Command to set the speed percentage for a fan platform entity."""

    command: Literal["fan_set_percentage"] = "fan_set_percentage"
    percentage: Optional[conint(ge=0, le=100)]  # type: ignore #TODO see if there is a way to make this work mypy


class FanSetPresetModeCommand(PlatformEntityCommand):
    """Command to set the preset mode for a fan platform entity."""

    command: Literal["fan_set_preset_mode"] = "fan_set_preset_mode"
    preset_mode: Optional[str]


class CoverOpenCommand(PlatformEntityCommand):
    """Command to open a cover platform entity."""

    command: Literal["cover_open"] = "cover_open"


class CoverCloseCommand(PlatformEntityCommand):
    """Command to open a cover platform entity."""

    command: Literal["cover_close"] = "cover_close"


class CoverSetPositionCommand(PlatformEntityCommand):
    """Command to set the position of a cover platform entity."""

    command: Literal["cover_set_position"] = "cover_set_position"
    position: int


class CoverStopCommand(PlatformEntityCommand):
    """Command to stop a cover platform entity."""

    command: Literal["cover_stop"] = "cover_stop"


class ClimateSetFanModeCommand(PlatformEntityCommand):
    """Command to set the fan mode for a climate platform entity."""

    command: Literal["climate_set_fan_mode"] = "climate_set_fan_mode"
    fan_mode: Optional[str]


class ClimateSetHvacModeCommand(PlatformEntityCommand):
    """Command to set the hvac mode for a climate platform entity."""

    command: Literal["climate_set_hvac_mode"] = "climate_set_hvac_mode"
    hvac_mode: Literal["heat_cool", "heat", "cool", "auto", "dry", "fan_only", "off"]


class ClimateSetPresetModeCommand(PlatformEntityCommand):
    """Command to set the preset mode for a climate platform entity."""

    command: Literal["climate_set_preset_mode"] = "climate_set_preset_mode"
    preset_mode: Optional[str]


class ClimateSetTemperatureCommand(PlatformEntityCommand):
    """Command to set the temperature for a climate platform entity."""

    command: Literal["climate_set_preset_mode"] = "climate_set_preset_mode"
    hvac_mode: Optional[
        Literal["heat_cool", "heat", "cool", "auto", "dry", "fan_only", "off"]
    ]
    temperature: Optional[float]
    target_temp_high: Optional[float]
    target_temp_low: Optional[float]


class ButtonPressCommand(PlatformEntityCommand):
    """Command to issue the press command for a button platform entity."""

    command: Literal["button_press"] = "button_press"


class AlarmControlPanelDisarmCommand(PlatformEntityCommand):
    """Command to disarm an alarm control panel platform entity."""

    command: Literal["alarm_control_panel_disarm"] = "alarm_control_panel_disarm"
    code: Optional[str]


class AlarmControlPanelArmHomeCommand(PlatformEntityCommand):
    """Command to arm an alarm control panel platform entity in home mode."""

    command: Literal["alarm_control_panel_arm_home"] = "alarm_control_panel_arm_home"
    code: Optional[str]


class AlarmControlPanelArmAwayCommand(PlatformEntityCommand):
    """Command to arm an alarm control panel platform entity in away mode."""

    command: Literal["alarm_control_panel_arm_away"] = "alarm_control_panel_arm_away"
    code: Optional[str]


class AlarmControlPanelArmNightCommand(PlatformEntityCommand):
    """Command to arm an alarm control panel platform entity in night mode."""

    command: Literal["alarm_control_panel_arm_night"] = "alarm_control_panel_arm_night"
    code: Optional[str]


class AlarmControlPanelTriggerCommand(PlatformEntityCommand):
    """Command to trigger the alarm for an alarm control panel platform entity."""

    command: Literal["alarm_control_panel_trigger"] = "alarm_control_panel_trigger"
    code: Optional[str]


class PlatformEntityRefreshStateCommand(PlatformEntityCommand):
    """Command to refresh the state of a platform entity."""

    command: Literal["platform_entity_refresh_state"] = "platform_entity_refresh_state"


class ClientListenCommand(Command):
    """Command to listen to events for a client."""

    command: Literal["client_listen"] = "client_listen"


class ClientListenRawZCLCommand(Command):
    """Command to listen to raw zcl events for a client."""

    command: Literal["client_listen_raw_zcl"] = "client_listen_raw_zcl"


class ClientDisconnectCommand(Command):
    """Command to disconnect a client."""

    command: Literal["client_disconnect"] = "client_disconnect"


class CommandResponse(BaseModel):
    """Command response model."""

    message_type: Literal["result"] = "result"
    message_id: int
    success: bool


class DefaultResponse(CommandResponse):
    """Get devices response."""

    command: Literal[
        "start_network",
        "stop_network",
        "remove_device",
        "stop_server",
        "light_turn_on",
        "light_turn_off",
        "switch_turn_on",
        "switch_turn_off",
        "lock_lock",
        "lock_unlock",
        "lock_set_user_lock_code",
        "lock_clear_user_lock_code",
        "lock_disable_user_lock_code",
        "lock_enable_user_lock_code",
        "fan_turn_on",
        "fan_turn_off",
        "fan_set_percentage",
        "fan_set_preset_mode",
        "cover_open",
        "cover_close",
        "cover_set_position",
        "cover_stop",
        "climate_set_fan_mode",
        "climate_set_hvac_mode",
        "climate_set_preset_mode",
        "climate_set_temperature",
        "button_press",
        "alarm_control_panel_disarm",
        "alarm_control_panel_arm_home",
        "alarm_control_panel_arm_away",
        "alarm_control_panel_arm_night",
        "alarm_control_panel_trigger",
        "select_select_option",
        "siren_turn_on",
        "siren_turn_off",
        "numbet_set_value",
        "platform_entity_refresh_state",
        "client_listen",
        "client_listen_raw_zcl",
        "client_disconnect",
        "reconfigure_device",
        "UpdateNetworkTopologyCommand",
    ]


class PermitJoiningResponse(CommandResponse):
    """Get devices response."""

    command: Literal["permit_joining"] = "permit_joining"
    duration: int


class GetDevicesResponse(CommandResponse):
    """Get devices response."""

    command: Literal["get_devices"] = "get_devices"
    devices: dict[str, Device]


class ReadClusterAttributesResponse(CommandResponse):
    """Read cluster attributes response."""

    command: Literal["read_cluster_attributes"] = "read_cluster_attributes"
    device: MinimalDevice
    cluster: MinimalCluster
    manufacturer_code: Optional[int]
    succeeded: dict[str, Any]
    failed: dict[str, Any]


class AttributeStatus(BaseModel):
    """Attribute status."""

    attribute: str
    status: str


class WriteClusterAttributeResponse(CommandResponse):
    """Write cluster attribute response."""

    command: Literal["write_cluster_attribute"] = "write_cluster_attribute"
    device: MinimalDevice
    cluster: MinimalCluster
    manufacturer_code: Optional[int]
    response: AttributeStatus


class GroupsResponse(CommandResponse):
    """Get groups response."""

    command: Literal["get_groups", "create_group"]
    groups: dict[int, Group]


class UpdateGroupResponse(CommandResponse):
    """Update group response."""

    command: Literal["create_group", "add_group_members", "remove_group_members"]
    group: Group


CommandResponses = Annotated[
    Union[
        DefaultResponse,
        GetDevicesResponse,
        GroupsResponse,
        PermitJoiningResponse,
        UpdateGroupResponse,
        ReadClusterAttributesResponse,
        WriteClusterAttributeResponse,
    ],
    Field(discriminator="command"),  # noqa: F821
]



"""Sensor platform for zhawss."""
from __future__ import annotations

import asyncio
import functools
import logging
import numbers
from typing import TYPE_CHECKING, Any, Final, Type, Union

from zhaws.server.decorators import periodic
from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_ANALOG_INPUT,
    CLUSTER_HANDLER_BASIC,
    CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT,
    CLUSTER_HANDLER_HUMIDITY,
    CLUSTER_HANDLER_ILLUMINANCE,
    CLUSTER_HANDLER_LEAF_WETNESS,
    CLUSTER_HANDLER_POWER_CONFIGURATION,
    CLUSTER_HANDLER_PRESSURE,
    CLUSTER_HANDLER_SMARTENERGY_METERING,
    CLUSTER_HANDLER_SOIL_MOISTURE,
    CLUSTER_HANDLER_TEMPERATURE,
    CLUSTER_HANDLER_THERMOSTAT,
)
from zhaws.server.zigbee.registries import SMARTTHINGS_HUMIDITY_CLUSTER

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.SENSOR)
CLUSTER_HANDLER_ST_HUMIDITY_CLUSTER = f"channel_0x{SMARTTHINGS_HUMIDITY_CLUSTER:04x}"

_LOGGER = logging.getLogger(__name__)

BATTERY_SIZES = {
    0: "No battery",
    1: "Built in",
    2: "Other",
    3: "AA",
    4: "AAA",
    5: "C",
    6: "D",
    7: "CR2",
    8: "CR123A",
    9: "CR2450",
    10: "CR2032",
    11: "CR1632",
    255: "Unknown",
}

CURRENT_HVAC_OFF: Final[str] = "off"
CURRENT_HVAC_HEAT: Final[str] = "heating"
CURRENT_HVAC_COOL: Final[str] = "cooling"
CURRENT_HVAC_DRY: Final[str] = "drying"
CURRENT_HVAC_IDLE: Final[str] = "idle"
CURRENT_HVAC_FAN: Final[str] = "fan"


class Sensor(PlatformEntity):
    """Representation of a zhawss sensor."""

    PLATFORM = Platform.SENSOR
    SENSOR_ATTR: Union[int, str, None] = None
    _REFRESH_INTERVAL = (30, 45)
    _decimals: int = 1
    _divisor: int = 1
    _multiplier: int | float = 1
    _unit: Union[str, None] = None

    @classmethod
    def create_platform_entity(
        cls: Type[Sensor],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Return a platform entity if it is a supported configuration, otherwise return None
        """
        cluster_handler = cluster_handlers[0]
        if cls.SENSOR_ATTR in cluster_handler.cluster.unsupported_attributes:
            return None

        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the sensor."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: ClusterHandler = cluster_handlers[0]
        self._cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )
        if self.should_poll:
            self._tracked_tasks.append(
                asyncio.create_task(
                    self._refresh(),
                    name=f"sensor_state_poller_{self.unique_id}_{self.__class__.__name__}",
                )
            )

    def get_state(self) -> dict:
        """Return the state for this sensor."""
        response = super().get_state()
        if self.SENSOR_ATTR is not None:
            raw_state = self._cluster_handler.cluster.get(self.SENSOR_ATTR)
            if raw_state is not None:
                raw_state = self.formatter(raw_state)
            response["state"] = raw_state
        return response

    def formatter(self, value: int) -> Union[int, float]:
        """Numeric pass-through formatter."""
        if self._decimals > 0:
            return round(
                float(value * self._multiplier) / self._divisor, self._decimals
            )
        return round(float(value * self._multiplier) / self._divisor)

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """handle attribute updates from the cluster handler."""
        self.maybe_send_state_changed_event()

    @periodic(_REFRESH_INTERVAL)
    async def _refresh(self) -> None:
        """Refresh the sensor."""
        await self.async_update()

    def to_json(self) -> dict:
        """Return a JSON representation of the sensor."""
        json = super().to_json()
        json["attribute"] = self.SENSOR_ATTR
        json["decimals"] = self._decimals
        json["divisor"] = self._divisor
        json["multiplier"] = self._multiplier
        json["unit"] = self._unit
        return json


@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ANALOG_INPUT,
    manufacturers="LUMI",
    models={"lumi.plug", "lumi.plug.maus01", "lumi.plug.mmeu01"},
    stop_on_match_group=CLUSTER_HANDLER_ANALOG_INPUT,
)
@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ANALOG_INPUT,
    manufacturers="Digi",
    stop_on_match_group=CLUSTER_HANDLER_ANALOG_INPUT,
)
class AnalogInput(Sensor):
    """Sensor that displays analog input values."""

    SENSOR_ATTR = "present_value"


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_POWER_CONFIGURATION)
class Battery(Sensor):
    """Battery sensor of power configuration cluster."""

    SENSOR_ATTR = "battery_percentage_remaining"

    @classmethod
    def create_platform_entity(
        cls: Type[Battery],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Unlike any other entity, PowerConfiguration cluster may not support
        battery_percent_remaining attribute, but zha-device-handlers takes care of it
        so create the entity regardless
        """
        if device.is_mains_powered:
            return None
        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    @staticmethod
    def formatter(value: int) -> int:
        """Return the state of the entity."""
        # per zcl specs battery percent is reported at 200% ¯\_(ツ)_/¯
        if not isinstance(value, numbers.Number) or value == -1:
            return value
        value = round(value / 2)
        return value

    def get_state(self) -> dict[str, Any]:
        """Return the state for battery sensors."""
        response = super().get_state()
        battery_size = self._cluster_handler.cluster.get("battery_size")
        if battery_size is not None:
            response["battery_size"] = BATTERY_SIZES.get(battery_size, "Unknown")
        battery_quantity = self._cluster_handler.cluster.get("battery_quantity")
        if battery_quantity is not None:
            response["battery_quantity"] = battery_quantity
        battery_voltage = self._cluster_handler.cluster.get("battery_voltage")
        if battery_voltage is not None:
            response["battery_voltage"] = round(battery_voltage / 10, 2)
        return response


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT)
class ElectricalMeasurement(Sensor):
    """Active power measurement."""

    SENSOR_ATTR = "active_power"
    _div_mul_prefix = "ac_power"

    @property
    def should_poll(self) -> bool:
        """Return True if we need to poll for state changes."""
        return True

    def get_state(self) -> dict[str, Any]:
        """Return the state for this sensor."""
        response = super().get_state()
        if self._cluster_handler.measurement_type is not None:
            response["measurement_type"] = self._cluster_handler.measurement_type

        max_attr_name = f"{self.SENSOR_ATTR}_max"
        if (max_v := self._cluster_handler.cluster.get(max_attr_name)) is not None:
            response[max_attr_name] = str(self.formatter(max_v))

        return response

    def formatter(self, value: int) -> Union[int, float]:
        """Return 'normalized' value."""
        multiplier = getattr(
            self._cluster_handler, f"{self._div_mul_prefix}_multiplier"
        )
        divisor = getattr(self._cluster_handler, f"{self._div_mul_prefix}_divisor")
        value = float(value * multiplier) / divisor
        if value < 100 and divisor > 1:
            return round(value, self._decimals)
        return round(value)

    async def async_update(self) -> None:
        """Retrieve latest state."""
        if not self.available:
            return
        await super().async_update()


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT)
class ElectricalMeasurementApparentPower(
    ElectricalMeasurement, id_suffix="apparent_power"
):
    """Apparent power measurement."""

    SENSOR_ATTR = "apparent_power"
    _div_mul_prefix = "ac_power"

    @property
    def should_poll(self) -> bool:
        """Poll indirectly by ElectricalMeasurementSensor."""
        return False


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT)
class ElectricalMeasurementRMSCurrent(ElectricalMeasurement, id_suffix="rms_current"):
    """RMS current measurement."""

    SENSOR_ATTR = "rms_current"
    _div_mul_prefix = "ac_current"

    @property
    def should_poll(self) -> bool:
        """Poll indirectly by ElectricalMeasurementSensor."""
        return False


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT)
class ElectricalMeasurementRMSVoltage(ElectricalMeasurement, id_suffix="rms_voltage"):
    """RMS Voltage measurement."""

    SENSOR_ATTR = "rms_voltage"
    _div_mul_prefix = "ac_voltage"

    @property
    def should_poll(self) -> bool:
        """Poll indirectly by ElectricalMeasurementSensor."""
        return False


@MULTI_MATCH(
    generic_ids=CLUSTER_HANDLER_ST_HUMIDITY_CLUSTER,
    stop_on_match_group=CLUSTER_HANDLER_HUMIDITY,
)
@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_HUMIDITY,
    stop_on_match_group=CLUSTER_HANDLER_HUMIDITY,
)
class Humidity(Sensor):
    """Humidity sensor."""

    SENSOR_ATTR = "measured_value"
    _divisor = 100


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_SOIL_MOISTURE)
class SoilMoisture(Sensor):
    """Soil Moisture sensor."""

    SENSOR_ATTR = "measured_value"
    _divisor = 100


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_LEAF_WETNESS)
class LeafWetness(Sensor):
    """Leaf Wetness sensor."""

    SENSOR_ATTR = "measured_value"
    _divisor = 100


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ILLUMINANCE)
class Illuminance(Sensor):
    """Illuminance Sensor."""

    SENSOR_ATTR = "measured_value"

    @staticmethod
    def formatter(value: int) -> float:
        """Convert illumination data."""
        return round(pow(10, ((value - 1) / 10000)), 1)


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_SMARTENERGY_METERING)
class SmartEnergyMetering(Sensor):
    """Metering sensor."""

    SENSOR_ATTR: Union[int, str] = "instantaneous_demand"

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the sensor."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._unit = self._cluster_handler.unit_of_measurement

    def formatter(self, value: int) -> Union[int, float]:
        """Pass through cluster handler formatter."""
        return self._cluster_handler.demand_formatter(value)

    def get_state(self) -> dict[str, Any]:
        """Return state for this sensor."""
        response = super().get_state()
        if self._cluster_handler.device_type is not None:
            response["device_type"] = self._cluster_handler.device_type
        if (status := self._cluster_handler.metering_status) is not None:
            response["status"] = str(status)[len(status.__class__.__name__) + 1 :]
        return response


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_SMARTENERGY_METERING)
class SmartEnergySummation(SmartEnergyMetering, id_suffix="summation_delivered"):
    """Smart Energy Metering summation sensor."""

    SENSOR_ATTR: Union[int, str] = "current_summ_delivered"

    def formatter(self, value: int) -> Union[int, float]:
        """Numeric pass-through formatter."""
        if self._cluster_handler.unit_of_measurement != 0:
            return self._cluster_handler.summa_formatter(value)

        cooked = (
            float(self._cluster_handler.multiplier * value)
            / self._cluster_handler.divisor
        )
        return round(cooked, 3)


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_PRESSURE)
class Pressure(Sensor):
    """Pressure sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_TEMPERATURE)
class Temperature(Sensor):
    """Temperature Sensor."""

    SENSOR_ATTR = "measured_value"
    _divisor = 100


@MULTI_MATCH(cluster_handler_names="carbon_dioxide_concentration")
class CarbonDioxideConcentration(Sensor):
    """Carbon Dioxide Concentration sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0
    _multiplier = 1e6


@MULTI_MATCH(cluster_handler_names="carbon_monoxide_concentration")
class CarbonMonoxideConcentration(Sensor):
    """Carbon Monoxide Concentration sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0
    _multiplier = 1e6


@MULTI_MATCH(generic_ids="channel_0x042e", stop_on_match_group="voc_level")
@MULTI_MATCH(cluster_handler_names="voc_level", stop_on_match_group="voc_level")
class VOCLevel(Sensor):
    """VOC Level sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0
    _multiplier = 1e6


@MULTI_MATCH(
    cluster_handler_names="voc_level",
    models="lumi.airmonitor.acn01",
    stop_on_match_group="voc_level",
)
class PPBVOCLevel(Sensor):
    """VOC Level sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0
    _multiplier = 1


@MULTI_MATCH(cluster_handler_names="formaldehyde_concentration")
class FormaldehydeConcentration(Sensor):
    """Formaldehyde Concentration sensor."""

    SENSOR_ATTR = "measured_value"
    _decimals = 0
    _multiplier = 1e6


@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class ThermostatHVACAction(Sensor, id_suffix="hvac_action"):
    """Thermostat HVAC action sensor."""

    @classmethod
    def create_platform_entity(
        cls: Type[ThermostatHVACAction],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Return entity if it is a supported configuration, otherwise return None
        """

        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    @property
    def _rm_rs_action(self) -> Union[str, None]:
        """Return the current HVAC action based on running mode and running state."""

        if (running_state := self._cluster_handler.running_state) is None:
            return None

        rs_heat = (
            self._cluster_handler.RunningState.Heat_State_On
            | self._cluster_handler.RunningState.Heat_2nd_Stage_On
        )
        if running_state & rs_heat:
            return CURRENT_HVAC_HEAT

        rs_cool = (
            self._cluster_handler.RunningState.Cool_State_On
            | self._cluster_handler.RunningState.Cool_2nd_Stage_On
        )
        if running_state & rs_cool:
            return CURRENT_HVAC_COOL

        running_state = self._cluster_handler.running_state
        if running_state and running_state & (
            self._cluster_handler.RunningState.Fan_State_On
            | self._cluster_handler.RunningState.Fan_2nd_Stage_On
            | self._cluster_handler.RunningState.Fan_3rd_Stage_On
        ):
            return CURRENT_HVAC_FAN

        running_state = self._cluster_handler.running_state
        if running_state and running_state & self._cluster_handler.RunningState.Idle:
            return CURRENT_HVAC_IDLE

        if self._cluster_handler.system_mode != self._cluster_handler.SystemMode.Off:
            return CURRENT_HVAC_IDLE
        return CURRENT_HVAC_OFF

    @property
    def _pi_demand_action(self) -> Union[str, None]:
        """Return the current HVAC action based on pi_demands."""

        heating_demand = self._cluster_handler.pi_heating_demand
        if heating_demand is not None and heating_demand > 0:
            return CURRENT_HVAC_HEAT
        cooling_demand = self._cluster_handler.pi_cooling_demand
        if cooling_demand is not None and cooling_demand > 0:
            return CURRENT_HVAC_COOL

        if self._cluster_handler.system_mode != self._cluster_handler.SystemMode.Off:
            return CURRENT_HVAC_IDLE
        return CURRENT_HVAC_OFF

    def get_state(self) -> dict:
        """Return the current HVAC action."""
        response = super().get_state()
        if (
            self._cluster_handler.pi_heating_demand is None
            and self._cluster_handler.pi_cooling_demand is None
        ):
            response["state"] = self._rm_rs_action
        response["state"] = self._pi_demand_action
        return response


@MULTI_MATCH(
    cluster_handler_names={CLUSTER_HANDLER_THERMOSTAT},
    manufacturers="Sinope Technologies",
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class SinopeHVACAction(ThermostatHVACAction):
    """Sinope Thermostat HVAC action sensor."""

    @property
    def _rm_rs_action(self) -> Union[str, None]:
        """Return the current HVAC action based on running mode and running state."""

        running_mode = self._cluster_handler.running_mode
        if running_mode == self._cluster_handler.RunningMode.Heat:
            return CURRENT_HVAC_HEAT
        if running_mode == self._cluster_handler.RunningMode.Cool:
            return CURRENT_HVAC_COOL

        running_state = self._cluster_handler.running_state
        if running_state and running_state & (
            self._cluster_handler.RunningState.Fan_State_On
            | self._cluster_handler.RunningState.Fan_2nd_Stage_On
            | self._cluster_handler.RunningState.Fan_3rd_Stage_On
        ):
            return CURRENT_HVAC_FAN
        if (
            self._cluster_handler.system_mode != self._cluster_handler.SystemMode.Off
            and running_mode == self._cluster_handler.SystemMode.Off
        ):
            return CURRENT_HVAC_IDLE
        return CURRENT_HVAC_OFF


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_BASIC)
class RSSISensor(Sensor, id_suffix="rssi"):
    """RSSI sensor for a device."""

    @classmethod
    def create_platform_entity(
        cls: Type[RSSISensor],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Return entity if it is a supported configuration, otherwise return None
        """
        key = f"{CLUSTER_HANDLER_BASIC}_{cls.unique_id_suffix}"
        if PLATFORM_ENTITIES.prevent_entity_creation(Platform.SENSOR, device.ieee, key):
            return None
        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    def get_state(self) -> dict:
        """Return the state of the sensor."""
        response = super().get_state()
        response["state"] = getattr(self.device.device, self.unique_id_suffix)  # type: ignore #TODO fix type hint
        return response

    @property
    def should_poll(self) -> bool:
        """Poll the sensor for current state."""
        return True


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_BASIC)
class LQISensor(RSSISensor, id_suffix="lqi"):
    """LQI sensor for a device."""


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_BASIC)
class LastSeenSensor(RSSISensor, id_suffix="last_seen"):
    """Last seen sensor for a device."""



"""Decorators for zhawss."""
from __future__ import annotations

import asyncio
from functools import wraps
import itertools
from random import uniform
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler

import zigpy.exceptions


def decorate_command(cluster_handler: ClusterHandler, command: Callable) -> Callable:
    """Wrap a cluster command to make it safe."""

    @wraps(command)
    async def wrapper(*args: Any, **kwds: Any) -> Any:
        try:
            result = await command(*args, **kwds)
            cluster_handler.debug(
                "executed '%s' command with args: '%s' kwargs: '%s' result: %s",
                command.__name__,
                args,
                kwds,
                result,
            )
            return result

        except (zigpy.exceptions.ZigbeeException, asyncio.TimeoutError) as ex:
            cluster_handler.debug(
                "command failed: '%s' args: '%s' kwargs '%s' exception: '%s'",
                command.__name__,
                args,
                kwds,
                str(ex),
            )
            return ex

    return wrapper


def retryable_request(
    delays: tuple = (1, 5, 10, 15, 30, 60, 120, 180, 360, 600, 900, 1800),
    raise_: bool = False,
) -> Callable:
    """Make a method with ZCL requests retryable.
    This adds delays keyword argument to function.
    len(delays) is number of tries.
    raise_ if the final attempt should raise the exception.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(
            cluster_handler: ClusterHandler, *args: Any, **kwargs: Any
        ) -> Any:

            exceptions = (zigpy.exceptions.ZigbeeException, asyncio.TimeoutError)
            try_count, errors = 1, []
            for delay in itertools.chain(delays, [None]):
                try:
                    return await func(cluster_handler, *args, **kwargs)
                except exceptions as ex:
                    errors.append(ex)
                    if delay:
                        delay = uniform(delay * 0.75, delay * 1.25)
                        cluster_handler.debug(
                            (
                                "%s: retryable request #%d failed: %s. "
                                "Retrying in %ss"
                            ),
                            func.__name__,
                            try_count,
                            ex,
                            round(delay, 1),
                        )
                        try_count += 1
                        await asyncio.sleep(delay)
                    else:
                        cluster_handler.warning(
                            "%s: all attempts have failed: %s", func.__name__, errors
                        )
                        if raise_:
                            raise

        return wrapper

    return decorator



"""Representation of a Zigbee endpoint for zhawss."""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Awaitable, Final

import zigpy
from zigpy.typing import EndpointType as ZigpyEndpointType

from zhaws.server.platforms import discovery
from zhaws.server.platforms.registries import Platform
from zhaws.server.zigbee import registries
from zhaws.server.zigbee.cluster import ClusterHandler
from zhaws.server.zigbee.cluster.general import MultistateInput

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import (
        ClientClusterHandler,
    )
    from zhaws.server.zigbee.device import Device

from zhaws.server.zigbee.decorators import CALLABLE_T

ATTR_DEVICE_TYPE: Final[str] = "device_type"
ATTR_PROFILE_ID: Final[str] = "profile_id"
ATTR_IN_CLUSTERS: Final[str] = "input_clusters"
ATTR_OUT_CLUSTERS: Final[str] = "output_clusters"

_LOGGER = logging.getLogger(__name__)


class Endpoint:
    """Endpoint for zhawss."""

    def __init__(self, zigpy_endpoint: ZigpyEndpointType, device: Device) -> None:
        """Initialize instance."""
        self._zigpy_endpoint: ZigpyEndpointType = zigpy_endpoint
        self._device: Device = device
        self._all_cluster_handlers: dict[str, ClusterHandler] = {}
        self._claimed_cluster_handlers: dict[str, ClusterHandler] = {}
        self._client_cluster_handlers: dict[str, ClientClusterHandler] = {}
        self._unique_id: str = f"{str(device.ieee)}-{zigpy_endpoint.endpoint_id}"

    @property
    def device(self) -> Device:
        """Return the device this endpoint belongs to."""
        return self._device

    @property
    def all_cluster_handlers(self) -> dict[str, ClusterHandler]:
        """All server cluster handlers of an endpoint."""
        return self._all_cluster_handlers

    @property
    def claimed_cluster_handlers(self) -> dict[str, ClusterHandler]:
        """Cluster handlers in use."""
        return self._claimed_cluster_handlers

    @property
    def client_cluster_handlers(self) -> dict[str, ClientClusterHandler]:
        """Return a dict of client cluster handlers."""
        return self._client_cluster_handlers

    @property
    def zigpy_endpoint(self) -> ZigpyEndpointType:
        """Return endpoint of zigpy device."""
        return self._zigpy_endpoint

    @property
    def id(self) -> int:
        """Return endpoint id."""
        return self._zigpy_endpoint.endpoint_id

    @property
    def unique_id(self) -> str:
        """Return the unique id for this endpoint."""
        return self._unique_id

    @property
    def zigbee_signature(self) -> tuple[int, dict[str, Any]]:
        """Get the zigbee signature for the endpoint this pool represents."""
        return (
            self.id,
            {
                ATTR_PROFILE_ID: f"0x{self._zigpy_endpoint.profile_id:04x}",
                ATTR_DEVICE_TYPE: f"0x{self._zigpy_endpoint.device_type:04x}"
                if self._zigpy_endpoint.device_type is not None
                else "",
                ATTR_IN_CLUSTERS: [
                    f"0x{cluster_id:04x}"
                    for cluster_id in sorted(self._zigpy_endpoint.in_clusters)
                ],
                ATTR_OUT_CLUSTERS: [
                    f"0x{cluster_id:04x}"
                    for cluster_id in sorted(self._zigpy_endpoint.out_clusters)
                ],
            },
        )

    @classmethod
    def new(cls, zigpy_endpoint: ZigpyEndpointType, device: Device) -> Endpoint:
        """Create new endpoint and populate cluster handlers."""
        endpoint = cls(zigpy_endpoint, device)
        endpoint.add_all_cluster_handlers()
        endpoint.add_client_cluster_handlers()
        discovery.PROBE.discover_entities(endpoint)
        return endpoint

    def add_all_cluster_handlers(self) -> None:
        """Create and add cluster handlers for all input clusters."""
        for cluster_id, cluster in self.zigpy_endpoint.in_clusters.items():
            cluster_handler_class = registries.CLUSTER_HANDLER_REGISTRY.get(
                cluster_id, ClusterHandler
            )
            _LOGGER.info(
                "Creating cluster handler for cluster id: %s class: %s",
                cluster_id,
                cluster_handler_class,
            )
            # really ugly hack to deal with xiaomi using the door lock cluster
            # incorrectly.
            if (
                hasattr(cluster, "ep_attribute")
                and cluster_id == zigpy.zcl.clusters.closures.DoorLock.cluster_id
                and cluster.ep_attribute == "multistate_input"
            ):
                cluster_handler_class = MultistateInput
            # end of ugly hack
            cluster_handler = cluster_handler_class(cluster, self)
            """ TODO
            if cluster_handler.name == CLUSTER_HANDLER_POWER_CONFIGURATION:
                if (
                    self._channels.power_configuration_ch
                    or self._channels.zha_device.is_mains_powered
                ):
                    # on power configuration channel per device
                    continue
                self._channels.power_configuration_ch = cluster_handler
            elif cluster_handler.name == CLUSTER_HANDLER_IDENTIFY:
                self._channels.identify_ch = channel
            """
            self._all_cluster_handlers[cluster_handler.id] = cluster_handler

    def add_client_cluster_handlers(self) -> None:
        """Create client cluster handlers for all output clusters if in the registry."""
        for (
            cluster_id,
            cluster_handler_class,
        ) in registries.CLIENT_CLUSTER_HANDLER_REGISTRY.items():
            cluster = self.zigpy_endpoint.out_clusters.get(cluster_id)
            if cluster is not None:
                cluster_handler = cluster_handler_class(cluster, self)
                self.client_cluster_handlers[cluster_handler.id] = cluster_handler

    async def async_initialize(self, from_cache: bool = False) -> None:
        """Initialize claimed cluster handlers."""
        await self._execute_handler_tasks("async_initialize", from_cache)

    async def async_configure(self) -> None:
        """Configure claimed cluster handlers."""
        await self._execute_handler_tasks("async_configure")

    async def _execute_handler_tasks(self, func_name: str, *args: Any) -> None:
        """Add a throttled cluster handler task and swallow exceptions."""

        async def _throttle(coro: Awaitable) -> None:
            async with self._device.semaphore:
                return await coro

        cluster_handlers = [
            *self.claimed_cluster_handlers.values(),
            *self.client_cluster_handlers.values(),
        ]
        tasks = [_throttle(getattr(ch, func_name)(*args)) for ch in cluster_handlers]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        for cluster_handler, outcome in zip(cluster_handlers, results):
            if isinstance(outcome, Exception):
                cluster_handler.warning(
                    "'%s' stage failed: %s", func_name, str(outcome)
                )
                continue
            cluster_handler.debug("'%s' stage succeeded", func_name)

    def async_new_entity(
        self,
        platform: Platform | str,
        entity_class: CALLABLE_T,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
    ) -> None:
        from zhaws.server.zigbee.device import DeviceStatus

        if self.device.status == DeviceStatus.INITIALIZED:
            return

        self.device.controller.server.data[platform].append(
            (entity_class, (unique_id, cluster_handlers, self, self.device))
        )

    def send_event(self, signal: dict[str, Any]) -> None:
        """Broadcast an event from this endpoint."""
        signal["endpoint"] = {
            "id": self.id,
            "unique_id": self.unique_id,
        }
        self.device.send_event(signal)

    def claim_cluster_handlers(self, cluster_handlers: list[ClusterHandler]) -> None:
        """Claim cluster handlers."""
        self.claimed_cluster_handlers.update({ch.id: ch for ch in cluster_handlers})

    def unclaimed_cluster_handlers(self) -> list[ClusterHandler]:
        """Return a list of available (unclaimed) cluster handlers."""
        claimed = set(self.claimed_cluster_handlers)
        available = set(self.all_cluster_handlers)
        return [
            self.all_cluster_handlers[cluster_id]
            for cluster_id in (available - claimed)
        ]



"""Constants for the cluster module."""

from typing import Final, Tuple

REPORT_CONFIG_ATTR_PER_REQ: Final[int] = 3
REPORT_CONFIG_MAX_INT: Final[int] = 900
REPORT_CONFIG_MAX_INT_BATTERY_SAVE: Final[int] = 10800
REPORT_CONFIG_MIN_INT: Final[int] = 30
REPORT_CONFIG_MIN_INT_ASAP: Final[int] = 1
REPORT_CONFIG_MIN_INT_IMMEDIATE: Final[int] = 0
REPORT_CONFIG_MIN_INT_OP: Final[int] = 5
REPORT_CONFIG_MIN_INT_BATTERY_SAVE: Final[int] = 3600
REPORT_CONFIG_RPT_CHANGE: Final[int] = 1
REPORT_CONFIG_DEFAULT: Tuple[int, int, int] = (
    REPORT_CONFIG_MIN_INT,
    REPORT_CONFIG_MAX_INT,
    REPORT_CONFIG_RPT_CHANGE,
)
REPORT_CONFIG_ASAP: Tuple[int, int, int] = (
    REPORT_CONFIG_MIN_INT_ASAP,
    REPORT_CONFIG_MAX_INT,
    REPORT_CONFIG_RPT_CHANGE,
)
REPORT_CONFIG_BATTERY_SAVE: Tuple[int, int, int] = (
    REPORT_CONFIG_MIN_INT_BATTERY_SAVE,
    REPORT_CONFIG_MAX_INT_BATTERY_SAVE,
    REPORT_CONFIG_RPT_CHANGE,
)
REPORT_CONFIG_IMMEDIATE: Tuple[int, int, int] = (
    REPORT_CONFIG_MIN_INT_IMMEDIATE,
    REPORT_CONFIG_MAX_INT,
    REPORT_CONFIG_RPT_CHANGE,
)
REPORT_CONFIG_OP: Tuple[int, int, int] = (
    REPORT_CONFIG_MIN_INT_OP,
    REPORT_CONFIG_MAX_INT,
    REPORT_CONFIG_RPT_CHANGE,
)
CLUSTER_READS_PER_REQ: Final[int] = 5

CLUSTER_HANDLER_ACCELEROMETER: Final[str] = "accelerometer"
CLUSTER_HANDLER_BINARY_INPUT: Final[str] = "binary_input"
CLUSTER_HANDLER_ANALOG_INPUT: Final[str] = "analog_input"
CLUSTER_HANDLER_ANALOG_OUTPUT: Final[str] = "analog_output"
CLUSTER_HANDLER_ATTRIBUTE: Final[str] = "attribute"
CLUSTER_HANDLER_BASIC: Final[str] = "basic"
CLUSTER_HANDLER_COLOR: Final[str] = "light_color"
CLUSTER_HANDLER_COVER: Final[str] = "window_covering"
CLUSTER_HANDLER_DOORLOCK: Final[str] = "door_lock"
CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT: Final[str] = "electrical_measurement"
CLUSTER_HANDLER_EVENT_RELAY: Final[str] = "event_relay"
CLUSTER_HANDLER_FAN: Final[str] = "fan"
CLUSTER_HANDLER_HUMIDITY: Final[str] = "humidity"
CLUSTER_HANDLER_SOIL_MOISTURE: Final[str] = "soil_moisture"
CLUSTER_HANDLER_LEAF_WETNESS: Final[str] = "leaf_wetness"
CLUSTER_HANDLER_IAS_ACE: Final[str] = "ias_ace"
CLUSTER_HANDLER_IAS_WD: Final[str] = "ias_wd"
CLUSTER_HANDLER_IDENTIFY: Final[str] = "identify"
CLUSTER_HANDLER_ILLUMINANCE: Final[str] = "illuminance"
CLUSTER_HANDLER_LEVEL: Final[str] = "level"
CLUSTER_HANDLER_MULTISTATE_INPUT: Final[str] = "multistate_input"
CLUSTER_HANDLER_OCCUPANCY: Final[str] = "occupancy"
CLUSTER_HANDLER_ON_OFF: Final[str] = "on_off"
CLUSTER_HANDLER_POWER_CONFIGURATION: Final[str] = "power"
CLUSTER_HANDLER_PRESSURE: Final[str] = "pressure"
CLUSTER_HANDLER_SHADE: Final[str] = "shade"
CLUSTER_HANDLER_SMARTENERGY_METERING: Final[str] = "smartenergy_metering"
CLUSTER_HANDLER_TEMPERATURE: Final[str] = "temperature"
CLUSTER_HANDLER_THERMOSTAT: Final[str] = "thermostat"
CLUSTER_HANDLER_ZDO: Final[str] = "zdo"
ZONE: Final[str] = "ias_zone"
CLUSTER_HANDLER_ZONE: Final[str] = ZONE



"""WS api for the switch platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server


class SwitchCommands(StrEnum):
    """Switch commands."""

    TURN_ON = "switch_turn_on"
    TURN_OFF = "switch_turn_off"


@decorators.websocket_command(platform_entity_command_schema(SwitchCommands.TURN_ON))
@decorators.async_response
async def turn_on(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn on the switch."""
    await execute_platform_entity_command(server, client, message, "async_turn_on")


@decorators.websocket_command(platform_entity_command_schema(SwitchCommands.TURN_OFF))
@decorators.async_response
async def turn_off(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn on the switch."""
    await execute_platform_entity_command(server, client, message, "async_turn_off")


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, turn_on)
    register_api_command(server, turn_off)



"""Models for platform things."""

from typing import Final, Literal, Optional

from zhaws.model import BaseEvent

STATE_CHANGED: Final[Literal["state_changed"]] = "state_changed"


class EntityStateChangedEvent(BaseEvent):
    """Event for when an entity state changes."""

    event_type: Literal["entity"] = "entity"
    event: Literal["state_changed"] = STATE_CHANGED
    platform: str
    unique_id: str
    device_ieee: Optional[str]
    endpoint_id: Optional[int]
    group_id: Optional[int]



"""Fan platform for zhawss."""
from __future__ import annotations

from abc import abstractmethod
import functools
import math
from typing import TYPE_CHECKING, Any, Final, Union

from zigpy.exceptions import ZigbeeException
from zigpy.zcl.clusters import hvac

from zhaws.server.platforms import BaseEntity, GroupEntity, PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_FAN
from zhaws.server.zigbee.group import Group

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.FAN)
GROUP_MATCH = functools.partial(PLATFORM_ENTITIES.group_match, Platform.FAN)

# Additional speeds in zigbee's ZCL
# Spec is unclear as to what this value means. On King Of Fans HBUniversal
# receiver, this means Very High.
PRESET_MODE_ON: Final[str] = "on"
# The fan speed is self-regulated
PRESET_MODE_AUTO: Final[str] = "auto"
# When the heated/cooled space is occupied, the fan is always on
PRESET_MODE_SMART: Final[str] = "smart"

SPEED_RANGE: Final = (1, 3)  # off is not included
PRESET_MODES_TO_NAME: Final[dict[int, str]] = {
    4: PRESET_MODE_ON,
    5: PRESET_MODE_AUTO,
    6: PRESET_MODE_SMART,
}

NAME_TO_PRESET_MODE = {v: k for k, v in PRESET_MODES_TO_NAME.items()}
PRESET_MODES = list(NAME_TO_PRESET_MODE)

DEFAULT_ON_PERCENTAGE: Final[int] = 50

ATTR_PERCENTAGE: Final[str] = "percentage"
ATTR_PRESET_MODE: Final[str] = "preset_mode"
SUPPORT_SET_SPEED: Final[int] = 1


def ranged_value_to_percentage(
    low_high_range: tuple[float, float], value: float
) -> int:
    """Given a range of low and high values convert a single value to a percentage.
    When using this utility for fan speeds, do not include 0 if it is off
    Given a low value of 1 and a high value of 255 this function
    will return:
    (1,255), 255: 100
    (1,255), 127: 50
    (1,255), 10: 4
    """
    offset = low_high_range[0] - 1
    return int(((value - offset) * 100) // states_in_range(low_high_range))


def percentage_to_ranged_value(
    low_high_range: tuple[float, float], percentage: int
) -> float:
    """Given a range of low and high values convert a percentage to a single value.
    When using this utility for fan speeds, do not include 0 if it is off
    Given a low value of 1 and a high value of 255 this function
    will return:
    (1,255), 100: 255
    (1,255), 50: 127.5
    (1,255), 4: 10.2
    """
    offset = low_high_range[0] - 1
    return states_in_range(low_high_range) * percentage / 100 + offset


def states_in_range(low_high_range: tuple[float, float]) -> float:
    """Given a range of low and high values return how many states exist."""
    return low_high_range[1] - low_high_range[0] + 1


def int_states_in_range(low_high_range: tuple[float, float]) -> int:
    """Given a range of low and high values return how many integer states exist."""
    return int(states_in_range(low_high_range))


class NotValidPresetModeError(ValueError):
    """Exception class when the preset_mode in not in the preset_modes list."""


class BaseFan(BaseEntity):
    """Base representation of a zhawss fan."""

    PLATFORM = Platform.FAN

    @property
    def preset_modes(self) -> list[str]:
        """Return the available preset modes."""
        return PRESET_MODES

    @property
    def supported_features(self) -> int:
        """Flag supported features."""
        return SUPPORT_SET_SPEED

    @property
    def speed_count(self) -> int:
        """Return the number of speeds the fan supports."""
        return int_states_in_range(SPEED_RANGE)

    async def async_turn_on(
        self,
        speed: str | None = None,
        percentage: int | None = None,
        preset_mode: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Turn the entity on."""
        if percentage is None:
            percentage = DEFAULT_ON_PERCENTAGE
        await self.async_set_percentage(percentage)

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn the entity off."""
        await self.async_set_percentage(0)

    async def async_set_percentage(self, percentage: int) -> None:
        """Set the speed percenage of the fan."""
        fan_mode = math.ceil(percentage_to_ranged_value(SPEED_RANGE, percentage))
        await self._async_set_fan_mode(fan_mode)

    async def async_set_preset_mode(self, preset_mode: str) -> None:
        """Set the preset mode for the fan."""
        if preset_mode not in self.preset_modes:
            raise NotValidPresetModeError(
                f"The preset_mode {preset_mode} is not a valid preset_mode: {self.preset_modes}"
            )
        await self._async_set_fan_mode(NAME_TO_PRESET_MODE[preset_mode])

    @abstractmethod
    async def _async_set_fan_mode(self, fan_mode: int) -> None:
        """Set the fan mode for the fan."""

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle state update from cluster handler."""
        self.maybe_send_state_changed_event()

    def to_json(self) -> dict:
        """Return a JSON representation of the binary sensor."""
        json = super().to_json()
        json["preset_modes"] = self.preset_modes
        json["supported_features"] = self.supported_features
        json["speed_count"] = self.speed_count
        return json


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_FAN)
class Fan(PlatformEntity, BaseFan):
    """Representation of a zhawss fan."""

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the fan."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._fan_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_FAN
        ]
        self._fan_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def percentage(self) -> Union[int, None]:
        """Return the current speed percentage."""
        if (
            self._fan_cluster_handler.fan_mode is None
            or self._fan_cluster_handler.fan_mode > SPEED_RANGE[1]
        ):
            return None
        if self._fan_cluster_handler.fan_mode == 0:
            return 0
        return ranged_value_to_percentage(
            SPEED_RANGE, self._fan_cluster_handler.fan_mode
        )

    @property
    def preset_mode(self) -> Union[str, None]:
        """Return the current preset mode."""
        return PRESET_MODES_TO_NAME.get(self._fan_cluster_handler.fan_mode)

    def async_set_state(self, attr_id: int, attr_name: str, value: Any) -> None:
        """Handle state update from cluster handler."""
        self.maybe_send_state_changed_event()

    async def _async_set_fan_mode(self, fan_mode: int) -> None:
        """Set the fan mode for the fan."""
        await self._fan_cluster_handler.async_set_speed(fan_mode)
        self.async_set_state(0, "fan_mode", fan_mode)

    def get_state(self) -> dict:
        """Return the state of the fan."""
        response = super().get_state()
        response.update(
            {
                "preset_mode": self.preset_mode,
                "percentage": self.percentage,
            }
        )
        return response


@GROUP_MATCH()
class FanGroup(GroupEntity, BaseFan):
    """Representation of a fan group."""

    def __init__(self, group: Group):
        """Initialize a fan group."""
        super().__init__(group)
        self._fan_channel = group.zigpy_group.endpoint[hvac.Fan.cluster_id]
        self._percentage = None
        self._preset_mode = None

    @property
    def percentage(self) -> int | None:
        """Return the current speed percentage."""
        return self._percentage

    @property
    def preset_mode(self) -> str | None:
        """Return the current preset mode."""
        return self._preset_mode

    def get_state(self) -> dict:
        """Return the state of the fan."""
        response = super().get_state()
        response.update(
            {
                "preset_mode": self.preset_mode,
                "percentage": self.percentage,
            }
        )
        return response

    async def _async_set_fan_mode(self, fan_mode: int) -> None:
        """Set the fan mode for the group."""
        try:
            await self._fan_channel.write_attributes({"fan_mode": fan_mode})
        except ZigbeeException as ex:
            self.error("Could not set fan mode: %s", ex)
        self.update()

    def update(self, _: Any = None) -> None:
        """Attempt to retrieve on off state from the fan."""
        self.debug("Updating fan group entity state")
        platform_entities = self._group.get_platform_entities(self.PLATFORM)
        all_entities = [entity.to_json() for entity in platform_entities]
        all_states = [entity["state"] for entity in all_entities]
        self.debug(
            "All platform entity states for group entity members: %s", all_states
        )

        self._available = any(entity.available for entity in platform_entities)
        percentage_states: list[dict] = [
            state for state in all_states if state.get(ATTR_PERCENTAGE)
        ]
        preset_mode_states: list[dict] = [
            state for state in all_states if state.get(ATTR_PRESET_MODE)
        ]

        if percentage_states:
            self._percentage = percentage_states[0][ATTR_PERCENTAGE]
            self._preset_mode = None
        elif preset_mode_states:
            self._preset_mode = preset_mode_states[0][ATTR_PRESET_MODE]
            self._percentage = None
        else:
            self._percentage = None
            self._preset_mode = None

        self.maybe_send_state_changed_event()



"""WS API for the cover platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.platforms.cover import ATTR_POSITION
from zhaws.server.websocket.api import decorators, register_api_command

# open, close, set position, stop


if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server


class CoverCommands(StrEnum):
    """Cover commands."""

    OPEN = "cover_open"
    CLOSE = "cover_close"
    SET_POSITION = "cover_set_position"
    STOP = "cover_stop"


@decorators.websocket_command(platform_entity_command_schema(CoverCommands.OPEN))
@decorators.async_response
async def open(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Open the cover."""
    await execute_platform_entity_command(server, client, message, "async_open_cover")


@decorators.websocket_command(platform_entity_command_schema(CoverCommands.CLOSE))
@decorators.async_response
async def close(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Close the cover."""
    await execute_platform_entity_command(server, client, message, "async_close_cover")


@decorators.websocket_command(
    platform_entity_command_schema(
        CoverCommands.SET_POSITION,
        {
            vol.Required(ATTR_POSITION): int,
        },
    )
)
@decorators.async_response
async def set_position(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Set the cover position."""
    await execute_platform_entity_command(
        server, client, message, "async_set_cover_position"
    )


@decorators.websocket_command(platform_entity_command_schema(CoverCommands.STOP))
@decorators.async_response
async def stop(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Stop the cover."""
    await execute_platform_entity_command(server, client, message, "async_stop_cover")


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, open)
    register_api_command(server, close)
    register_api_command(server, set_position)
    register_api_command(server, stop)



"""Siren platform for zhawss."""
from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, Any, Final, Union

import zigpy.types as t
from zigpy.zcl.clusters.security import IasWd as WD

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_IAS_WD

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.SIREN)
DEFAULT_DURATION: Final[int] = 5  # seconds
WARNING_DEVICE_MODE_STOP: Final[int] = 0
WARNING_DEVICE_MODE_BURGLAR: Final[int] = 1
WARNING_DEVICE_MODE_FIRE: Final[int] = 2
WARNING_DEVICE_MODE_EMERGENCY: Final[int] = 3
WARNING_DEVICE_MODE_POLICE_PANIC: Final[int] = 4
WARNING_DEVICE_MODE_FIRE_PANIC: Final[int] = 5
WARNING_DEVICE_MODE_EMERGENCY_PANIC: Final[int] = 6

WARNING_DEVICE_STROBE_NO: Final[int] = 0
WARNING_DEVICE_STROBE_YES: Final[int] = 1

WARNING_DEVICE_SOUND_LOW: Final[int] = 0
WARNING_DEVICE_SOUND_MEDIUM: Final[int] = 1
WARNING_DEVICE_SOUND_HIGH: Final[int] = 2
WARNING_DEVICE_SOUND_VERY_HIGH: Final[int] = 3

WARNING_DEVICE_STROBE_LOW: Final[int] = 0x00
WARNING_DEVICE_STROBE_MEDIUM: Final[int] = 0x01
WARNING_DEVICE_STROBE_HIGH: Final[int] = 0x02
WARNING_DEVICE_STROBE_VERY_HIGH: Final[int] = 0x03

WARNING_DEVICE_SQUAWK_MODE_ARMED: Final[int] = 0
WARNING_DEVICE_SQUAWK_MODE_DISARMED: Final[int] = 1

ATTR_TONE: Final[str] = "tone"

ATTR_AVAILABLE_TONES: Final[str] = "available_tones"
ATTR_DURATION: Final[str] = "duration"
ATTR_VOLUME_LEVEL: Final[str] = "volume_level"

SUPPORT_TURN_ON: Final[int] = 1
SUPPORT_TURN_OFF: Final[int] = 2
SUPPORT_TONES: Final[int] = 4
SUPPORT_VOLUME_SET: Final[int] = 8
SUPPORT_DURATION: Final[int] = 16


class Strobe(t.enum8):  # type: ignore #TODO fix type
    """Strobe enum."""

    No_Strobe = 0x00
    Strobe = 0x01


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_WD)
class Siren(PlatformEntity):
    """Representation of a zhawss siren."""

    PLATFORM = Platform.SIREN

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the siren."""
        self._attr_supported_features = (
            SUPPORT_TURN_ON
            | SUPPORT_TURN_OFF
            | SUPPORT_DURATION
            | SUPPORT_VOLUME_SET
            | SUPPORT_TONES
        )
        self._attr_available_tones: Union[
            list[Union[int, str]], dict[int, str], None
        ] = {
            WARNING_DEVICE_MODE_BURGLAR: "Burglar",
            WARNING_DEVICE_MODE_FIRE: "Fire",
            WARNING_DEVICE_MODE_EMERGENCY: "Emergency",
            WARNING_DEVICE_MODE_POLICE_PANIC: "Police Panic",
            WARNING_DEVICE_MODE_FIRE_PANIC: "Fire Panic",
            WARNING_DEVICE_MODE_EMERGENCY_PANIC: "Emergency Panic",
        }
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: ClusterHandler = cluster_handlers[0]
        self._attr_is_on: bool = False
        self._off_listener: asyncio.TimerHandle | None = None

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn on siren."""
        if self._off_listener:
            self._off_listener.cancel()
            self._off_listener = None
        tone_cache = self._cluster_handler.data_cache.get(
            WD.Warning.WarningMode.__name__
        )
        siren_tone = (
            tone_cache.value
            if tone_cache is not None
            else WARNING_DEVICE_MODE_EMERGENCY
        )
        siren_duration = DEFAULT_DURATION
        level_cache = self._cluster_handler.data_cache.get(
            WD.Warning.SirenLevel.__name__
        )
        siren_level = (
            level_cache.value if level_cache is not None else WARNING_DEVICE_SOUND_HIGH
        )
        strobe_cache = self._cluster_handler.data_cache.get(Strobe.__name__)
        should_strobe = (
            strobe_cache.value if strobe_cache is not None else Strobe.No_Strobe
        )
        strobe_level_cache = self._cluster_handler.data_cache.get(
            WD.StrobeLevel.__name__
        )
        strobe_level = (
            strobe_level_cache.value
            if strobe_level_cache is not None
            else WARNING_DEVICE_STROBE_HIGH
        )
        if (duration := kwargs.get(ATTR_DURATION)) is not None:
            siren_duration = duration
        if (tone := kwargs.get(ATTR_TONE)) is not None:
            siren_tone = tone
        if (level := kwargs.get(ATTR_VOLUME_LEVEL)) is not None:
            siren_level = int(level)
        await self._cluster_handler.issue_start_warning(
            mode=siren_tone,
            warning_duration=siren_duration,
            siren_level=siren_level,
            strobe=should_strobe,
            strobe_duty_cycle=50 if should_strobe else 0,
            strobe_intensity=strobe_level,
        )
        self._attr_is_on = True
        self._off_listener = asyncio.get_running_loop().call_later(
            siren_duration, self.async_set_off
        )
        self.maybe_send_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn off siren."""
        await self._cluster_handler.issue_start_warning(
            mode=WARNING_DEVICE_MODE_STOP, strobe=WARNING_DEVICE_STROBE_NO
        )
        self._attr_is_on = False
        self.maybe_send_state_changed_event()

    def async_set_off(self) -> None:
        """Set is_on to False and write HA state."""
        self._attr_is_on = False
        if self._off_listener:
            self._off_listener.cancel()
            self._off_listener = None
        self.maybe_send_state_changed_event()

    def to_json(self) -> dict:
        json = super().to_json()
        json[ATTR_AVAILABLE_TONES] = self._attr_available_tones
        json["supported_features"] = self._attr_supported_features
        return json

    def get_state(self) -> dict:
        """Get the state of the siren."""
        response = super().get_state()
        response["state"] = self._attr_is_on
        return response
